from .dgp import dgp
from .gp import gp
from .emulation import emulator
from .kernel_class import kernel, combine
from .likelihood_class import Poisson, Hetero, NegBin
from .linkgp import container, lgp
from .synthetic import path
from .utils import write, read, summary




import multiprocess.context as ctx
import platform
import numpy as np
from pathos.multiprocessing import ProcessingPool as Pool
import psutil 
import copy

class gp:
    """
    Class that for Gaussian process emulation.

    Args:
        X (ndarray): a numpy 2d-array where each row is an input data point and 
            each column is an input dimension.
        Y (ndarray): a numpy 2d-array with only one column and each row being an input data point.
        kernel (class): a :class:`.kernel` class that specifies the features of the GP. 
    """

    def __init__(self, X, Y, kernel):
        self.X=X
        self.Y=Y
        if (self.Y).ndim==1 or X.ndim==1:
            raise Exception('The input and output data have to be numpy 2d-arrays.')
        self.kernel=kernel
        self.initialize()

    def initialize(self):
        """Assign input/output data to the kernel for training.
        """
        if self.kernel.input_dim is not None:
            self.kernel.input=copy.deepcopy(self.X[:,self.kernel.input_dim])
        else:
            self.kernel.input=copy.deepcopy(self.X)
            self.kernel.input_dim=copy.deepcopy(np.arange(np.shape(self.X)[1]))
        if self.kernel.connect is not None:
            if len(np.intersect1d(self.kernel.connect,self.kernel.input_dim))!=0:
                raise Exception('The local input and global input should not have any overlap. Change input_dim or connect so they do not have any common indices.')
            self.kernel.global_input=copy.deepcopy(self.X[:,self.kernel.connect])
        self.kernel.output=copy.deepcopy(self.Y)

    def train(self):
        """Train the GP model.
        """
        self.kernel.maximise()
        self.kernel.compute_stats()

    def export(self):
        """Export the trained GP.
        """
        final_struct=copy.deepcopy(self.kernel)
        return [final_struct]

    def ppredict(self,x,method='mean_var',sample_size=50,chunk_num=None,core_num=None):
        """Implement parallel predictions from the trained GP model.

        Args:
            x, method, sample_size: see descriptions of the method :meth:`.gp.predict`.
            chunk_num (int, optional): the number of chunks that the testing input array **x** will be divided into. 
                Defaults to `None`. If not specified, the number of chunks is set to **core_num**. 
            core_num (int, optional): the number of cores/workers to be used. Defaults to `None`. If not specified, 
                the number of cores is set to ``(max physical cores available - 1)``.

        Returns:
            Same as the method :meth:`.gp.predict`.
        """
        if platform.system()=='Darwin':
            ctx._force_start_method('forkserver')
        if core_num==None:
            core_num=psutil.cpu_count(logical = False)-1
        if chunk_num==None:
            chunk_num=core_num
        if chunk_num<core_num:
            core_num=chunk_num
        f=lambda x: self.predict(*x) 
        z=np.array_split(x,chunk_num)
        with Pool(core_num) as pool:
            res = pool.map(f, [[x, method, sample_size] for x in z])
            pool.close()
            pool.join()
            pool.clear()
        if method == 'mean_var':
            return tuple(np.concatenate(worker) for worker in zip(*res))
        elif method == 'sampling':
            return np.concatenate(res)

    def predict(self,x,method='mean_var',sample_size=50):
        """Implement predictions from the trained GP model.

        Args:
            x (ndarray): a numpy 2d-array where each row is an input testing data point and 
                each column is an input dimension.
            method (str, optional): the prediction approach: mean-variance (`mean_var`) or sampling 
                (`sampling`) approach. Defaults to `mean_var`.
            sample_size (int, optional): the number of samples to draw from the predictive distribution of
                 GP if **method** = '`sampling`'. Defaults to `50`.

        Returns:
            tuple_or_ndarray: if the argument **method** = '`mean_var`', a tuple is returned:
                the tuple contains two numpy 2d-arrays, one for the predictive means 
                and another for the predictive variances. Each array has only one column with its rows 
                corresponding to testing positions;

            if the argument **method** = '`sampling`', a numpy 2d-array is returned:
                the array has its rows corresponding to testing positions and columns corresponding to
                sample_size number of samples drawn from the predictive distribution of GP;
        """
        if x.ndim==1:
            raise Exception('The testing input has to be a numpy 2d-array')
        M=len(x)
        overall_global_test_input=x
        if self.kernel.connect is not None:
            z_k_in=overall_global_test_input[:,self.kernel.connect]
        else:
            z_k_in=None
        if method=='mean_var':
            mu,sigma2=self.kernel.gp_prediction(x=overall_global_test_input[:,self.kernel.input_dim],z=z_k_in)
            return mu.reshape(-1,1), sigma2.reshape(-1,1)
        elif method=='sampling':
            mu,sigma2=self.kernel.gp_prediction(x=overall_global_test_input[:,self.kernel.input_dim],z=z_k_in)
            samples=np.random.normal(mu,np.sqrt(sigma2),size=(sample_size,M)).T
            return samples


import numpy as np
from numpy.random import randn, uniform, standard_t
from math import sqrt, pi
from scipy.optimize import minimize, Bounds
from scipy.linalg import cho_solve, pinvh
from scipy.spatial.distance import pdist, squareform
from .functions import gp, link_gp, pdist_matern_one, pdist_matern_multi, pdist_matern_coef, fod_exp, Z_fct

class kernel:
    """
    Class that defines the GPs in the DGP hierarchy.

    Args:
        length (ndarray): a numpy 1d-array, whose length equals to:

                1. either one if the lengthscales in the kernel function are assumed same across input dimensions; or
                2. the total number of input dimensions, which is the sum of the number of feeding GPs 
                   in the last layer (defined by the argument **input_dim**) and the number of connected global
                   input dimensions (defined by the argument **connect**), if the lengthscales in the kernel function 
                   are assumed different across input dimensions.
        scale (float, optional): the variance of a GP. Defaults to `1`.
        nugget (float, optional): the nugget term of a GP. Defaults to `1e-6`.
        name (str, optional): kernel function to be used. Either `sexp` for squared exponential kernel or
            `matern2.5` for Matern2.5 kernel. Defaults to `sexp`.
        prior_name (str, optional): prior options for the lengthscales and nugget term. Either gamma (`ga`) or inverse gamma (`inv_ga`) distribution for 
            the lengthscales and nugget term. Set `None` to disable the prior. Defaults to `ga`.
        prior_coef (ndarray, optional): a numpy 1d-array that contains two values specifying the shape and rate 
            parameters of gamma prior, shape and scale parameters of inverse gamma prior. Defaults to ``np.array([1.6,0.3])``.
        nugget_est (bool, optional): set to `True` to estimate nugget term or to `False` to fix the nugget term as specified
            by the argument **nugget**. If set to `True`, the value set to the argument **nugget** is used as the initial
            value. Defaults to `False`.
        scale_est (bool, optional): set to `True` to estimate the variance or to `False` to fix the variance as specified
            by the argument **scale**. Defaults to `False`.
        input_dim (ndarray, optional): a numpy 1d-array that contains either

                1. the indices of GPs in the feeding layer whose outputs feed into the GP; or
                2. the indices of dimensions in the global input if the GP is in the first layer. 
            When set to `None`, 
            
                1. all outputs from GPs in the feeding layer; or 
                2. all global input dimensions feed into the GP. 
            Defaults to `None`.
        connect (ndarray, optional): a numpy 1d-array that contains the indices of dimensions in the global
            input connecting to the GP as additional input dimensions to the input obtained from the output of
            GPs in the feeding layer (as determined by the argument **input_dim**). When set to `None`, no global input
            connection is implemented. Defaults to `None`. When the kernel class is used in GP/DGP emulators for linked
            emulation and some input dimensions to the computer models are not connected to some feeding computer models, 
            set **connect** to a 1d-array of indices of these external global input dimensions, and accordingly, set 
            **input_dim** to a 1d-array of indices of the remaining input dimensions that are connected to the feeding 
            computer models.                   

    Attributes:
        type (str): identifies that the kernel is a GP.
        g (function): a function giving the log probability density function of gamma or inverse gamma distribution 
            ignoring the constant part.
        gfod (function): a function giving the first order derivative of **g** with respect to the log-transformed 
            lengthscales and nugget. 
        para_path (ndarray): a numpy 2d-array that contains the trace of model parameters. Each row is a 
            parameter estimate produced by one SEM iteration. The model parameters in each row are ordered as 
            follow: ``np.array([scale estimate, lengthscale estimate (whose length>=1), nugget estimate])``.
        global_input (ndarray): a numpy 2d-array that contains the connect global input dimensions determined 
            by the argument **connect**. The value of the attribute is assigned during the initialisation of 
            :class:`.dgp` class. If **connect** is set to `None`, this attribute is also `None`. 
        input (ndarray): a numpy 2d-array (each row as a data point and each column as a data dimension) that 
            contains the input training data (according to the argument **input_dim**) to the GP. The value of 
            this attribute is assigned during the initialisation of :class:`.dgp` class. 
        output (ndarray): a numpy 2d-array with only one column that contains the output training data to the GP.
            The value of this attribute is assigned during the initialisation of :class:`.dgp` class.
        rep (ndarray): a numpy 1d-array used to re-construct repetitions in the data according to the repetitions 
            in the global input, i.e., rep is assigned during the initialisation of :class:`.dgp` class if one input position 
            has multiple outputs. Otherwise, it is `None`. Defaults to `None`. 
        Rinv (ndarray): a numpy 2d-array that stores the inversion of correlation matrix. Defaults to `None`.
        Rinv_y (ndarray): a numpy 2d-array that stores the product of correlation matrix inverse and the output Y. Defaults to `None`.
        rff (bool): indicates weather random Fourier features are used. Defaults to `None`.
        D (int): the dimension of input data to the GP node. Defaults to `None`.
        M (int): the number of features in random Fourier approximation. Defaults to `None`.
        W (ndarray): a 2d-array (M, D) sampled to construct the Fourier approximation to the kernel matrix. Defaults to `None`.
        b (ndarray): a 1d-array (D,) sampled to construct the Fourier approximation to the kernel matrix. Defaults to `None`.
    """

    def __init__(self, length, scale=1., nugget=1e-6, name='sexp', prior_name='ga', prior_coef=np.array([1.6,0.3]), nugget_est=False, scale_est=False, input_dim=None, connect=None):
        self.type='gp'
        self.length=length
        self.scale=np.atleast_1d(scale)
        self.nugget=np.atleast_1d(nugget)
        self.name=name
        self.prior_name=prior_name
        self.prior_coef=prior_coef
        if self.prior_name=='ga':
            self.g=lambda x: (self.prior_coef[0]-1)*np.log(x)-self.prior_coef[1]*x
            self.gfod=lambda x: (self.prior_coef[0]-1)-self.prior_coef[1]*x
        elif self.prior_name=='inv_ga':
            self.g=lambda x: -(self.prior_coef[0]+1)*np.log(x)-self.prior_coef[1]/x
            self.gfod=lambda x: -(self.prior_coef[0]+1)+self.prior_coef[1]/x
        self.nugget_est=nugget_est
        self.scale_est=scale_est
        self.input_dim=input_dim
        self.connect=connect
        self.para_path=np.atleast_2d(np.concatenate((self.scale,self.length,self.nugget)))
        self.global_input=None
        self.input=None
        self.output=None
        self.rep=None
        self.Rinv=None
        self.Rinv_y=None
        self.rff=None
        self.D=None
        self.M=None
        self.W=None
        self.b=None

    def sample_basis(self):
        """Sample **W** and **b** to construct random Fourier approximations to correlation matrices.
        """
        if self.name=='sexp':
            self.W=sqrt(2)*randn(self.M,self.D)
        elif self.name=='matern2.5':
            self.W=standard_t(5,size=(self.M,self.D))
        self.b=uniform(0,2*pi,size=self.M)

    def log_t(self):
        """Log transform the model parameters (lengthscales and nugget).

        Returns:
            ndarray: a numpy 1d-array of log-transformed model parameters
        """
        if self.nugget_est:
            log_theta=np.log(np.concatenate((self.length,self.nugget)))
        else:
            log_theta=np.log(self.length)
        return log_theta

    def update(self,log_theta):
        """Update the model parameters (scale, lengthscales and nugget).

        Args:
            log_theta (ndarray): optimised numpy 1d-array of log-transformed lengthscales and nugget.
        """
        theta=np.exp(log_theta)
        if self.nugget_est:
            self.length=theta[0:-1]
            self.nugget=theta[[-1]]
        else:
            self.length=theta
        if self.scale_est:
            K=self.k_matrix()
            L=np.linalg.cholesky(K)
            YKinvY=(self.output).T@cho_solve((L, True), self.output, check_finite=False)
            new_scale=YKinvY/len(self.output)
            self.scale=new_scale.flatten()
            
    def k_matrix(self,fod_eval=False):
        """Compute the correlation matrix and/or first order derivatives of the correlation matrix wrt log-transformed lengthscales and nugget.
        
        Args:
            fod_eval (bool): indicates if the gradient information is also computed along with the correlation
                matrix. Defaults to `False`. 

        Returns:
            ndarray_or_tuple: 
                1. If **fod_eval** = `False`, a numpy 2d-array *K* is returned as the correlation matrix.
                2. If **fod_eval** = `True`, a tuple is returned. It includes *K* and fod, a numpy 3d-array that contains the first order derivatives of the correlation matrix 
                   wrt log-transformed lengthscales and nugget. The length of the array equals to the total number 
                   of model parameters (i.e., the total number of lengthscales and nugget).
        """
        n=len(self.input)
        if self.global_input is not None:
            X=np.concatenate((self.input, self.global_input),1)
        else:
            X=self.input
        X_l=X/self.length
        if self.name=='sexp':
            dists = pdist(X_l, metric="sqeuclidean")
            K = squareform(np.exp(-dists))
            if fod_eval:
                if len(self.length)==1:
                    fod=np.expand_dims(squareform(2*dists)*K,axis=0)
                else:
                    fod=fod_exp(X_l,K)
        elif self.name=='matern2.5':
            if fod_eval:
                K=squareform(np.exp(-np.sqrt(5)*pdist(X_l, metric="minkowski",p=1)))
                if len(self.length)==1:
                    coef1, coef2 = pdist_matern_one(X_l)
                else:
                    coef1, coef2 = pdist_matern_multi(X_l)
                K*=coef1
                fod=coef2*K
            else:
                K=np.exp(-np.sqrt(5)*pdist(X_l, metric="minkowski",p=1))
                K*=pdist_matern_coef(X_l)
                K=squareform(K)
        if fod_eval and self.nugget_est:
            nugget_fod=np.expand_dims(self.nugget*np.eye(n),0)
            fod=np.concatenate((fod,nugget_fod),axis=0)
        np.fill_diagonal(K, 1+self.nugget)
        if fod_eval:
            return K, fod
        else:
            return K
    
    def log_prior(self):
        """Compute the value of log priors specified to the lengthscales and nugget. 

        Returns:
            ndarray: a numpy 1d-array giving the sum of log priors of the lengthscales and nugget. 
        """
        lp=np.sum(self.g(self.length),keepdims=True)
        if self.nugget_est:
            lp+=self.g(self.nugget)
        return lp

    def log_prior_fod(self):
        """Compute the first order derivatives of log priors wrt the log-transformed lengthscales and nugget.

        Returns:
            ndarray: a numpy 1d-array (whose length equal to the total number of lengthscales and nugget)
            giving the first order derivatives of log priors wrt the log-transformed lengthscales and nugget.
        """
        fod=self.gfod(self.length)
        if self.nugget_est:
            fod=np.concatenate((fod,self.gfod(self.nugget)))
        return fod

    def llik(self,x):
        """Compute the negative log-likelihood function of the GP.

        Args:
            x (ndarray): a numpy 1d-array that contains the values of log-transformed model parameters: 
                log-transformed lengthscales followed by the log-transformed nugget. 

        Returns:
            ndarray: a numpy 1d-array giving negative log-likelihood.
        """
        self.update(x)
        n=len(self.output)
        K=self.k_matrix()
        L=np.linalg.cholesky(K)
        logdet=2*np.sum(np.log(np.abs(np.diag(L))))
        YKinvY=(self.output).T@cho_solve((L, True), self.output, check_finite=False)
        if self.scale_est:
            scale=YKinvY/n
            neg_llik=0.5*(logdet+n*np.log(scale))
        else:
            neg_llik=0.5*(logdet+YKinvY/self.scale) 
        neg_llik=neg_llik.flatten()
        if self.prior_name!=None:
            neg_llik=neg_llik-self.log_prior()
        return neg_llik
    
    def llik_rff(self,x):
        """Compute the negative log-likelihood function of the GP under RFF.

        Args:
            x (ndarray): a numpy 1d-array that contains the values of log-transformed model parameters: 
                log-transformed lengthscales followed by the log-transformed nugget. 

        Returns:
            ndarray: a numpy 1d-array giving negative log-likelihood.
        """
        self.update(x)
        n=len(self.output)
        if self.connect is not None:
            X=np.concatenate((self.input,self.global_input),1)
        else:
            X=self.input
        Z=Z_fct(X,self.W,self.b,self.length,self.M)
        cov=np.dot(Z.T,Z)+self.nugget*np.identity(self.M)
        L=np.linalg.cholesky(cov)
        logdet=2*np.sum(np.log(np.abs(np.diag(L))))
        Zt_y=np.dot(Z.T,self.output)
        quad=np.dot(self.output.T,self.output)-np.sum(Zt_y*cho_solve((L, True), Zt_y, check_finite=False))
        if self.scale_est:
            scale=quad/(n*self.nugget)
            neg_llik=0.5*(logdet+n*np.log(scale)+(n-self.M)*np.log(self.nugget))
        else:
            neg_llik=0.5*(logdet+quad/(self.scale*self.nugget)+(n-self.M)*np.log(self.nugget)) 
        neg_llik=neg_llik.flatten()
        if self.prior_name!=None:
            neg_llik=neg_llik-self.log_prior()
        return neg_llik

    def llik_der(self,x):
        """Compute first order derivatives of the negative log-likelihood function wrt log-transformed model parameters.

        Args:
            x (ndarray): a numpy 1d-array that contains the values of log-transformed model parameters: 
                log-transformed lengthscales followed by the log-transformed nugget.

        Returns:
            ndarray: a numpy 1d-array (whose length equal to the total number of lengthscales and nugget)
            that contains first order derivatives of the negative log-likelihood function wrt log-transformed 
            lengthscales and nugget.
        """
        self.update(x)
        n=len(self.output)
        K,Kt=self.k_matrix(fod_eval=True)
        KinvKt=np.linalg.solve(K,Kt)
        tr_KinvKt=np.trace(KinvKt,axis1=1, axis2=2)
        L=np.linalg.cholesky(K)
        KinvY=cho_solve((L, True), self.output, check_finite=False)
        YKinvKtKinvY=((self.output).T@KinvKt@KinvY).flatten()
        P1=-0.5*tr_KinvKt
        P2=0.5*YKinvKtKinvY
        if self.scale_est:
            YKinvY=(self.output).T@KinvY
            scale=(YKinvY/n).flatten()
            neg_St=-P1-P2/scale
        else:
            neg_St=-P1-P2/self.scale
        if self.prior_name!=None:
            neg_St=neg_St-self.log_prior_fod()
        return neg_St

    def log_likelihood_func(self):
        cov=self.scale*self.k_matrix()
        L=np.linalg.cholesky(cov)
        logdet=2*np.sum(np.log(np.abs(np.diag(L))))
        quad=(self.output).T@cho_solve((L, True), self.output, check_finite=False)
        llik=-0.5*(logdet+quad)
        return llik

    def log_likelihood_func_rff(self):
        """Compute Gaussian log-likelihood function using random Fourier features (RFF).
        """
        if self.connect is not None:
            X=np.concatenate((self.input,self.global_input),1)
        else:
            X=self.input
        Z=Z_fct(X,self.W,self.b,self.length,self.M)
        cov=np.dot(Z.T,Z)+self.nugget*np.identity(self.M)
        L=np.linalg.cholesky(cov)
        logdet=2*np.sum(np.log(np.abs(np.diag(L))))
        Zt_y=np.dot(Z.T,self.output)
        quad=-np.sum(Zt_y*cho_solve((L, True), Zt_y, check_finite=False))/(self.scale*self.nugget)
        #quad=-(Zt_y.T@cho_solve((L, True), Zt_y, check_finite=False))/(self.scale*self.nugget)
        llik=-0.5*(logdet+quad)
        return llik

    def maximise(self, method='L-BFGS-B'):
        """Optimise and update model parameters by minimising the negative log-likelihood function.

        Args:
            method (str, optional): optimisation algorithm. Defaults to `L-BFGS-B`.
        """
        initial_theta_trans=self.log_t()
        if self.nugget_est:
            if self.rff:
                lb=np.concatenate((-5.*np.ones(len(initial_theta_trans)-1),np.log([1e-8])))
                ub=5.*np.ones(len(initial_theta_trans))
                bd=Bounds(lb, ub)
                _ = minimize(self.llik_rff, initial_theta_trans, method=method, bounds=bd, options={'maxiter': 100, 'maxfun': 125})
            else:
                lb=np.concatenate((-np.inf*np.ones(len(initial_theta_trans)-1),np.log([1e-8])))
                ub=np.inf*np.ones(len(initial_theta_trans))
                bd=Bounds(lb, ub)
                _ = minimize(self.llik, initial_theta_trans, method=method, jac=self.llik_der, bounds=bd, options={'maxiter': 100, 'maxfun': 125})
        else:
            if self.rff:
                lb=-5.*np.ones(len(initial_theta_trans))
                ub=5.*np.ones(len(initial_theta_trans))
                bd=Bounds(lb, ub)
                _ = minimize(self.llik_rff, initial_theta_trans, method=method, bounds=bd, options={'maxiter': 100, 'maxfun': 125})
            else:
                _ = minimize(self.llik, initial_theta_trans, method=method, jac=self.llik_der, options={'maxiter': 100, 'maxfun': 125})
        self.add_to_path()
        
    def add_to_path(self):
        """Add updated model parameter estimates to the class attribute **para_path**.
        """
        para=np.concatenate((self.scale,self.length,self.nugget))
        self.para_path=np.vstack((self.para_path,para))

    def gp_prediction(self,x,z):
        """Make GP predictions. 

        Args:
            x (ndarray): a numpy 2d-array that contains the input testing data (whose rows correspond to testing
                data points and columns correspond to testing data dimensions) with the number of columns same as 
                the **input** attribute.
            z (ndarray): a numpy 2d-array that contains additional input testing data (with the same number of 
                columns of the **global_input** attribute) from the global testing input if the argument **connect** 
                is not `None`. Set to None if the argument **connect** is `None`. 

        Returns:
            tuple: a tuple of two 1d-arrays giving the means and variances at the testing input data positions. 
        """
        m,v=gp(x,z,self.input,self.global_input,self.Rinv,self.Rinv_y,self.scale,self.length,self.nugget,self.name)
        return m,v

    def linkgp_prediction(self,m,v,z,nb_parallel):
        """Make linked GP predictions. 

        Args:
            m (ndarray): a numpy 2d-array that contains predictive means of testing outputs from the GPs in the last 
                layer. The number of rows equals to the number of testing positions and the number of columns equals to the 
                length of the argument **input_dim**. If the argument **input_dim** is `None`, then the number of columns equals 
                to the number of GPs in the last layer.
            v (ndarray): a numpy 2d-array that contains predictive variances of testing outputs from the GPs in the last 
                layer. It has the same shape of **m**.
            z (ndarray): a numpy 2d-array that contains additional input testing data (with the same number of 
                columns of the **global_input** attribute) from the global testing input if the argument **connect** 
                is not `None`. Set to `None` if the argument **connect** is `None`. 
            nb_parallel (bool): whether to use *Numba*'s multi-threading to accelerate the predictions.

        Returns:
            tuple: a tuple of two 1d-arrays giving the means and variances at the testing input data positions (that are 
            represented by predictive means and variances).
        """
        m,v=link_gp(m,v,z,self.input,self.global_input,self.Rinv,self.Rinv_y,self.scale,self.length,self.nugget,self.name,nb_parallel)
        return m,v

    def linkgp_prediction_full(self,m,v,m_z,v_z,z,nb_parallel):
        """Make linked GP predictions with additional input also generated by GPs/DGPs. 

        Args:
            m (ndarray): a numpy 2d-array that contains predictive means of testing outputs from the GPs in the last 
                layer. The number of rows equals to the number of testing positions and the number of columns equals to the 
                length of the argument **input_dim**. If the argument **input_dim** is `None`, then the number of columns equals 
                to the number of GPs in the last layer.
            v (ndarray): a numpy 2d-array that contains predictive variances of testing outputs from the GPs in the last 
                layer. It has the same shape of **m**.
            m_z (ndarray): a numpy 2d-array that contains predictive means of additional input testing data from GPs.
            v_z (ndarray): a numpy 2d-array that contains predictive variances of additional input testing data from GPs.
            z (ndarray): a numpy 2d-array that contains additional input testing data from the global testing input that are
                not from GPs. Set to `None` if the argument **connect** is None. 
            nb_parallel (bool): whether to use *Numba*'s multi-threading to accelerate the predictions.

        Returns:
            tuple: a tuple of two 1d-arrays giving the means and variances at the testing input data positions (that are 
            represented by predictive means and variances).
        """
        m=np.concatenate((m,m_z),axis=1)
        v=np.concatenate((v,v_z),axis=1)
        idx1=np.arange(np.shape(m_z)[1])
        idx2=np.arange(np.shape(m_z)[1],np.shape(self.global_input)[1])
        overall_input=np.concatenate((self.input,self.global_input[:,idx1]),axis=1)
        m,v=link_gp(m,v,z,overall_input,self.global_input[:,idx2],self.Rinv,self.Rinv_y,self.scale,self.length,self.nugget,self.name,nb_parallel)
        return m,v

    def compute_stats(self):
        """Compute and store key statistics for the GP predictions
        """
        R=self.k_matrix()
        #U, s, Vh = np.linalg.svd(R)
        #self.Rinv=Vh.T@np.diag(s**-1)@U.T
        #L=np.linalg.cholesky(R)
        #self.Rinv_y=cho_solve((L, True), self.output, check_finite=False)
        self.Rinv=pinvh(R,check_finite=False)
        self.Rinv_y=np.dot(self.Rinv,self.output)

def combine(*layers):
    """Combine layers into one list as a DGP structure.

    Args:
        layers (list): a sequence of lists, each of which contains the GPs (defined by the :class:`.kernel` class) in that layer.

    Returns:
        list: a list of layers defining the DGP structure.
    """
    all_layer=[]
    for layer in layers:
        all_layer.append(layer)
    return all_layer


import multiprocess.context as ctx
import platform
import numpy as np
from pathos.multiprocessing import ProcessingPool as Pool
import psutil
from .imputation import imputer
import copy

class container:
    """
    Class to contain the trained GP or DGP emulator of a computer model for linked (D)GP emulation.

    Args:
        structure (list): a list that contains the trained structure of GP or DGP of a computer model. For GP, 
            this is the list exported from the :meth:`.export` method of the :class:`.gp` class. For DGP, this is the list exported 
            from the :meth:`.estimate` of the :class:`.dgp` class.
        local_input_idx (ndarray): a numpy 1d-array that specifies the indices of outputs (a 2d-array) 
            produced by all models in the feeding layer that are input to the model emulated by the GP or DGP
            represented by the **structure** argument. The indices should be ordered in such a way that the extracted
            output from the feeding layer is sorted in the same order as the training input used for the GP/DGP 
            emulation of the computer model that the **structure** argument represents. When the model is in the first
            layer, **local_input_idx** gives the indices of its input in the global testing input set, see :meth:`.lgp.predict`
            for descriptions of the global testing input set. Defaults to `None`. When the
            argument is `None`, one needs to set its value using the :meth:`.set_local_input`. 
    """
    def __init__(self, structure, local_input_idx=None):
        if len(structure)==1:
            self.type='gp'
            self.structure=structure[0]
        else:
            self.type='dgp'
            self.structure=structure
            self.imp=imputer(self.structure)
            self.imp.sample(burnin=50)
        self.local_input_idx=local_input_idx

    def set_local_input(self, idx):
        """Set the **local_input_idx** argument if it is not set when the class is constructed.
        """
        self.local_input_idx=idx

class lgp:
    """
    Class to store a system of GP and DGP emulators for predictions. 

    Args:
        all_layer (list): a list contains *L* (the number of layers of a systems of computer models) sub-lists, 
            each of which represents a layer and contains the GP/DGP emulators of computer models represented by 
            the :class:`.container` class. The sub-lists are placed in the list in the same order of the specified computer
            model system.
        N (int): the number of imputation to produce the predictions. Increase the value to account for more 
            imputation uncertainties. If the system consists only GP emulators, **N** is set to `1` automatically. 
            Defaults to `50`.
        nb_parallel (bool, optional): whether to use *Numba*'s multi-threading to accelerate the predictions. Defaults to `False`.
    """
    def __init__(self, all_layer, N=50, nb_parallel=False):
        self.nb_parallel=nb_parallel
        self.L=len(all_layer)
        self.all_layer=all_layer
        self.num_model=[]
        for l in range(1,self.L):
            self.num_model.append(len(all_layer[l]))
        if np.sum([[cont.type=='dgp' for cont in all_layer[l]] for l in range(self.L)])==0:
            N=1
        self.all_layer_set=[]
        for _ in range(N):
            one_imputation=[]
            for l in range(self.L):
                layer=[]
                for cont in all_layer[l]:
                    if cont.type=='gp':
                        layer.append(copy.deepcopy(cont))
                    elif cont.type=='dgp':
                        (cont.imp).sample()
                        (cont.imp).key_stats()
                        layer.append(copy.deepcopy(cont))
                one_imputation.append(layer)
            self.all_layer_set.append(one_imputation)

    def set_nb_parallel(self,nb_parallel):
        """Set 'self.nb_parallel' to the bool value given by **nb_parallel**. This method is useful to change **self.nb_parallel**
            when the :class:`.lgp` class has already been built.
        """
        self.nb_parallel=nb_parallel
    
    def ppredict(self,x,method='mean_var',full_layer=False,sample_size=50,chunk_num=None,core_num=None):
        """Implement parallel predictions from the trained DGP model.

        Args:
            x, method, full_layer, sample_size: see descriptions of the method :meth:`.lgp.predict`.
            chunk_num (int, optional): the number of chunks that the testing input array **x** will be divided into. 
                Defaults to `None`. If not specified, the number of chunks is set to **core_num**. 
            core_num (int, optional): the number of cores/workers to be used. Defaults to `None`. If not specified, 
                the number of cores is set to ``(max physical cores available - 1)``.

        Returns:
            Same as the method `predict`.
        """
        if platform.system()=='Darwin':
            ctx._force_start_method('forkserver')
        if core_num==None:
            core_num=psutil.cpu_count(logical = False)-1
        if chunk_num==None:
            chunk_num=core_num
        if chunk_num<core_num:
            core_num=chunk_num
        f=lambda x: self.predict(*x) 
        if isinstance(x, list):
            if len(x)!=self.L:
                raise Exception('When test input is given as a list, it must contain global inputs to the all layers (even with no global inputs to internal layers). Set None as the global input to the internal models if they have no global inputs.')
            else:
                for l in range(self.L):
                    if l==0:
                        z=[[element] for element in np.array_split(x[l],chunk_num)]
                    else:
                        z_l=x[l]
                        z_m=[[]]*chunk_num
                        for m in range(len(z_l)):
                            if z_l[m]==None:
                                z_m=[i+j for i, j in zip(z_m,[[None]]*chunk_num)]
                            else:
                                z_m=[i+[j] for i, j in zip(z_m,np.array_split(z_l[m],chunk_num))]
                        z=[i+[j] for i,j in zip(z,z_m)]
        elif not isinstance(x, list):
            z=np.array_split(x,chunk_num)
        with Pool(core_num) as pool:
            res = pool.map(f, [[x, method, full_layer, sample_size] for x in z])
        if method == 'mean_var':
            if full_layer:
                combined_res=[]
                for comp in zip(*res):
                    combined_comp=[]
                    for layer in zip(*comp):
                        combined_comp.append(list(np.concatenate(worker) for worker in zip(*list(layer))))
                    combined_res.append(combined_comp)
                return tuple(combined_res)
            else:
                combined_res=[]
                for comp in zip(*res):
                    combined_res.append(list(np.concatenate(workers) for workers in zip(*list(comp))))
                return tuple(combined_res)
        elif method == 'sampling':
            if full_layer:
                combined_res=[]
                for layer in zip(*res):
                    combined_res.append(list(np.concatenate(workers,axis=1) for workers in zip(*list(layer))))
                return combined_res
            else:
                return list(np.concatenate(worker,axis=1) for worker in zip(*res))

    def predict(self,x,method='mean_var',full_layer=False,sample_size=50):
        """Implement predictions from the linked (D)GP model.

        Args:
            x (ndarray_or_list): a numpy 2d-array or a list.
            
                1. If **x** is a 2d-array, it is the global testing input set to the computer models in the first layer where 
                   each rows are input testing data points and columns are input dimensions across all computer models in 
                   the first layer of the system. In this case, it is assumed that **x** is the only global input to the computer 
                   system, i.e., there are no external global input to computer models in layers other than the first layer.
                2. If **x** is a list, it has *L* (the number of layers of a systems of computer models) elements. The first element
                   is a numpy 2d-array that represents the global testing input set to the computer models in the first layer. 
                   The remaining *L-1* elements are *L-1* sub-lists, each of which contains a number (same as the number of computer
                   models in the corresponding layer) of numpy 2d-arrays (rows being testing points and columns being input 
                   dimensions) that represent the external global testing input to the computer models in the corresponding layer. 
                   The order of 2d-arrays in each sub-list must be the same order of the emulators placed in the corresponding layer
                   of **all_layer** argument to :class:`.lgp` class. If there is no external global input to a certain computer model of the 
                   system, set `None` in the corresponding sub-list (i.e., layer) of **x**.   
            method (str, optional): the prediction approach: mean-variance (`mean_var`) or sampling 
                (`sampling`) approach. Defaults to `mean_var`.
            full_layer (bool, optional): whether to output the predictions from all GP/DGP emulators in the system. Defaults to `False`.
            sample_size (int, optional): the number of samples to draw for each given imputation if **method** = '`sampling`'.
                Defaults to `50`.
            
        Returns:
            tuple_or_list: 
                if the argument **method** = '`mean_var`', a tuple is returned:
                    1. If **full_layer** = `False`, the tuple contains two lists, one for the predictive means 
                       and another for the predictive variances. Each list contains a number (same number of computer models in the
                       final layer of the system) of numpy 2d-arrays. Each 2d-array has its rows corresponding to global testing 
                       positions and columns corresponding to GP/DGP (or DGP+likelihood) output dimensions of the associated computer
                       model in the final layer;
                    2. If **full_layer** = `True`, the tuple contains two lists, one for the predictive means 
                       and another for the predictive variances. Each list contains *L* (i.e., the number of layers of the emulated system) 
                       sub-lists. Each sub-list represents a layer and contains a number (same number of computer models in the corresponding 
                       layer of the system) of numpy 2d-arrays. Each array has its rows corresponding to global testing positions and columns 
                       corresponding to GP/DGP (or DGP+likelihood in case of the final layer) output dimensions of the associated computer
                       model in the corresponding layer.
                if the argument **method** = '`sampling`', a list is returned:
                    1. If **full_layer** = `False`, the list contains a number (same number of computer models in the final layer of the system) of numpy 
                       3d-arrays. Each array corresponds to a computer model in the final layer, and has its 0-axis corresponding to the output 
                       dimensions of the GP/DGP (or DGP+likelihood) emulators, 1-axis corresponding to global testing positions, and 2-axis 
                       corresponding to samples of size **N** * **sample_size**;
                    2. If **full_layer** = `True`, the list contains *L* (i.e., the number of layers of the emulated system) sub-lists. Each sub-list 
                       represents a layer and contains a number (same number of computer models in the corresponding layer of the system) of 
                       numpy 3d-arrays. Each array corresponds to a computer model in the associated layer, and has its 0-axis corresponding 
                       to the output dimensions of the GP/DGP (or DGP+likelihood in case of the final layer) emulators, 1-axis corresponding 
                       to global testing positions, and 2-axis corresponding to samples of size **N** * **sample_size**.

        """
        if isinstance(x, list) and len(x)!=self.L:
            raise Exception('When test input is given as a list, it must contain global inputs to the all layers (even with no global inputs to internal layers). Set None as the global input to the internal models if they have no global inputs.')
        elif not isinstance(x, list):
            if x.ndim==1:
                raise Exception('The testing input has to be a numpy 2d-array.')
            x=[x]
            for num in self.num_model:
                x.append([None]*num) 
        if method=='mean_var':
            sample_size=1
        mean_pred=[]
        variance_pred=[]
        if method=='sampling':
            sample_pred=[]
        for s in range(len(self.all_layer_set)):
            one_imputed_all_layer=self.all_layer_set[s]
            if full_layer:
                if method=='mean_var':
                    mean_pred_oneN=[]
                    variance_pred_oneN=[]
                elif method=='sampling':
                    sample_pred_oneN=[]
            for l in range(self.L):
                layer=one_imputed_all_layer[l]
                n_model=len(layer)
                if l==self.L-1:
                    m_last_layer,v_last_layer=[], []
                    if method=='sampling':
                        sample_last_layer=[]
                else:
                    m_l,v_l=[], []
                    if method=='sampling':
                        sample_l=[]
                if l==0:
                    for k in range(n_model):
                        model=layer[k]
                        input_lk=x[l][:,model.local_input_idx]
                        if model.type=='gp':
                            m_lk, v_lk = self.gp_pred(input_lk,None,None,None,model.structure,self.nb_parallel)
                        elif model.type=='dgp':
                            _, _, m_lk, v_lk = self.dgp_pred(input_lk,None,None,None,model.structure,self.nb_parallel)
                        m_l.append(m_lk)
                        v_l.append(v_lk)
                        if method=='sampling' and full_layer:
                            row_num, col_num = np.shape(m_lk)
                            sample_lk=np.random.normal(m_lk,np.sqrt(v_lk),size=(sample_size, row_num, col_num))
                            sample_l.append(sample_lk.transpose(2,1,0))
                    if full_layer:
                        if method=='mean_var':
                            mean_pred_oneN.append(m_l)
                            variance_pred_oneN.append(v_l)
                        elif method=='sampling':
                            sample_pred_oneN.append(sample_l)
                    m_l_next, v_l_next=np.concatenate(m_l,axis=1), np.concatenate(v_l,axis=1)
                elif l==self.L-1:
                    for k in range(n_model):
                        model=layer[k]
                        external_input_lk=x[l][k]
                        m_input_lk, v_input_lk = m_l_next[:,model.local_input_idx], v_l_next[:,model.local_input_idx]
                        if model.type=='gp':
                            m_lk, v_lk=self.gp_pred(None,m_input_lk,v_input_lk,external_input_lk,model.structure,self.nb_parallel)
                            if method=='sampling':
                                row_num, col_num = np.shape(m_lk)
                                sample_lk=np.random.normal(m_lk,np.sqrt(v_lk),size=(sample_size, row_num, col_num)).transpose(2,1,0)
                        elif model.type=='dgp':
                            m_one_before_lk, v_one_before_lk, m_lk, v_lk = self.dgp_pred(None,m_input_lk,v_input_lk,external_input_lk,model.structure,self.nb_parallel)
                            if method=='sampling':
                                row_num, col_num = np.shape(m_lk)
                                sample_lk=np.empty((col_num, row_num, sample_size))
                                for count, kernel in enumerate(model.structure[-1]):
                                    if kernel.type=='gp':
                                        sample_lk[count,]=np.random.normal(m_lk[:,[count]],np.sqrt(v_one_before_lk[:,[count]]),size=(row_num,sample_size))
                                    elif kernel.type=='likelihood':
                                        dgp_sample=np.random.normal(m_one_before_lk,np.sqrt(v_one_before_lk),size=(sample_size, m_one_before_lk.shape[0], m_one_before_lk.shape[1]))
                                        sample_lk[count,]=np.array([kernel.sampling(dgp_sample[:,:,[kernel.input_dim]][i]) for i in range(sample_size)]).T
                        if method=='mean_var':                     
                            m_last_layer.append(m_lk)
                            v_last_layer.append(v_lk)
                        elif method=='sampling':
                            sample_last_layer.append(sample_lk)
                    if full_layer:
                        if method=='mean_var':
                            mean_pred_oneN.append(m_last_layer)
                            variance_pred_oneN.append(v_last_layer)
                        elif method=='sampling':
                            sample_pred_oneN.append(sample_last_layer)
                else:
                    for k in range(n_model):
                        model=layer[k]
                        external_input_lk=x[l][k]
                        m_input_lk, v_input_lk = m_l_next[:,model.local_input_idx], v_l_next[:,model.local_input_idx]
                        if model.type=='gp':
                            m_lk, v_lk = self.gp_pred(None,m_input_lk,v_input_lk,external_input_lk,model.structure,self.nb_parallel)
                        elif model.type=='dgp':
                            _, _, m_lk, v_lk = self.dgp_pred(None,m_input_lk,v_input_lk,external_input_lk,model.structure,self.nb_parallel)
                        m_l.append(m_lk)
                        v_l.append(v_lk)
                        if method=='sampling' and full_layer:
                            row_num, col_num = np.shape(m_lk)
                            sample_lk=np.random.normal(m_lk,np.sqrt(v_lk),size=(sample_size, row_num, col_num))
                            sample_l.append(sample_lk.transpose(2,1,0))
                    if full_layer:
                        if method=='mean_var':
                            mean_pred_oneN.append(m_l)
                            variance_pred_oneN.append(v_l)
                        elif method=='sampling':
                            sample_pred_oneN.append(sample_l)
                    m_l_next, v_l_next=np.concatenate(m_l,axis=1), np.concatenate(v_l,axis=1)
            if full_layer:
                if method=='mean_var':
                    mean_pred.append(mean_pred_oneN)
                    variance_pred.append(variance_pred_oneN)
                elif method=='sampling':
                    sample_pred.append(sample_pred_oneN)
            else:
                if method=='mean_var':
                    mean_pred.append(m_last_layer)
                    variance_pred.append(v_last_layer)
                elif method=='sampling':
                    sample_pred.append(sample_last_layer)
        if method=='mean_var':
            if full_layer:
                mu=[[np.mean(i,axis=0) for i in zip(*case_m)] for case_m in zip(*mean_pred)]
                sigma2=[[np.mean(np.square(i)+j,axis=0)-np.mean(i,axis=0)**2 for i,j in zip(zip(*case_m),zip(*case_v))] for case_m, case_v in zip(zip(*mean_pred),zip(*variance_pred))]
            else:
                mu=[np.mean(i,axis=0) for i in zip(*mean_pred)]
                sigma2=[np.mean(np.square(i)+j,axis=0)-np.mean(i,axis=0)**2 for i, j in zip(zip(*mean_pred),zip(*variance_pred))]
            return mu, sigma2
        elif method=='sampling':
            if full_layer:
                samples=[[np.concatenate(i,axis=2) for i in zip(*case_s)] for case_s in zip(*sample_pred)]
            else:
                samples=[np.concatenate(i,axis=2) for i in zip(*sample_pred)]
            return samples

    @staticmethod
    def gp_pred(x,m,v,z,structure,nb_parallel):
        """Compute predictive mean and variance from a GP emulator when the testing input is either deterministic or normally distributed.
        """
        if x is None:
            m,v=structure.linkgp_prediction(m=m,v=v,z=z,nb_parallel=nb_parallel)
        else:
            m,v=structure.gp_prediction(x=x,z=z)
        return m.reshape(-1,1),v.reshape(-1,1)
    
    @staticmethod
    def dgp_pred(x,m,v,z,structure,nb_parallel):
        """Compute predictive mean and variance from a DGP (DGP+likelihood) emulator when the testing input is either deterministic or normally distributed.
        """
        if x is None:
            M=len(m)
        else:
            M=len(x)
        L=len(structure)
        for l in range(L):
            layer=structure[l]
            n_kerenl=len(layer)
            if l==L-1:
                likelihood_gp_mean=np.empty((M,n_kerenl))
                likelihood_gp_var=np.empty((M,n_kerenl))
            else:
                overall_test_output_mean=np.empty((M,n_kerenl))
                overall_test_output_var=np.empty((M,n_kerenl))
            if l==0:
                for k in range(n_kerenl):
                    kernel=layer[k]
                    if x is None:
                        m_k,v_k=kernel.linkgp_prediction(m=m,v=v,z=z,nb_parallel=nb_parallel)
                    else:
                        m_k,v_k=kernel.gp_prediction(x=x,z=z)
                    overall_test_output_mean[:,k],overall_test_output_var[:,k]=m_k,v_k
                overall_test_input_mean,overall_test_input_var=overall_test_output_mean,overall_test_output_var
            elif l==L-1:
                for k in range(n_kerenl):
                    kernel=layer[k]
                    m_k_in,v_k_in=overall_test_input_mean[:,kernel.input_dim],overall_test_input_var[:,kernel.input_dim]
                    if kernel.type=='gp':
                        if kernel.connect is not None:
                            if x is None:
                                D=np.shape(m)[1]
                                idx1,idx2=kernel.connect[kernel.connect<=(D-1)],kernel.connect[kernel.connect>(D-1)]
                                if idx1.size==0:
                                    m_k,v_k=kernel.linkgp_prediction(m=m_k_in,v=v_k_in,z=z[:,idx2-D],nb_parallel=nb_parallel)
                                elif idx2.size==0:
                                    m_k,v_k=kernel.linkgp_prediction_full(m=m_k_in,v=v_k_in,m_z=m[:,idx1],v_z=v[:,idx1],z=None,nb_parallel=nb_parallel)
                                else:
                                    m_k,v_k=kernel.linkgp_prediction_full(m=m_k_in,v=v_k_in,m_z=m[:,idx1],v_z=v[:,idx1],z=z[:,idx2-D],nb_parallel=nb_parallel)
                            else:
                                m_k,v_k=kernel.linkgp_prediction(m=m_k_in,v=v_k_in,z=x[:,kernel.connect],nb_parallel=nb_parallel)
                        else:
                            m_k,v_k=kernel.linkgp_prediction(m=m_k_in,v=v_k_in,z=None,nb_parallel=nb_parallel)
                        likelihood_gp_mean[:,k],likelihood_gp_var[:,k]=m_k,v_k
                    elif kernel.type=='likelihood':
                        m_k,v_k=kernel.prediction(m=m_k_in,v=v_k_in)
                        likelihood_gp_mean[:,k],likelihood_gp_var[:,k]=m_k,v_k
            else:
                for k in range(n_kerenl):
                    kernel=layer[k]
                    m_k_in,v_k_in=overall_test_input_mean[:,kernel.input_dim],overall_test_input_var[:,kernel.input_dim]
                    if kernel.connect is not None:
                        if x is None:
                            D=np.shape(m)[1]
                            idx1,idx2=kernel.connect[kernel.connect<=(D-1)],kernel.connect[kernel.connect>(D-1)]
                            if idx1.size==0:
                                m_k,v_k=kernel.linkgp_prediction(m=m_k_in,v=v_k_in,z=z[:,idx2-D],nb_parallel=nb_parallel)
                            elif idx2.size==0:
                                m_k,v_k=kernel.linkgp_prediction_full(m=m_k_in,v=v_k_in,m_z=m[:,idx1],v_z=v[:,idx1],z=None,nb_parallel=nb_parallel)
                            else:
                                m_k,v_k=kernel.linkgp_prediction_full(m=m_k_in,v=v_k_in,m_z=m[:,idx1],v_z=v[:,idx1],z=z[:,idx2-D],nb_parallel=nb_parallel)
                        else:
                            m_k,v_k=kernel.linkgp_prediction(m=m_k_in,v=v_k_in,z=x[:,kernel.connect],nb_parallel=nb_parallel)
                    else:
                        m_k,v_k=kernel.linkgp_prediction(m=m_k_in,v=v_k_in,z=None,nb_parallel=nb_parallel)
                    overall_test_output_mean[:,k],overall_test_output_var[:,k]=m_k,v_k
                overall_test_input_mean,overall_test_input_var=overall_test_output_mean,overall_test_output_var
        return overall_test_input_mean, overall_test_input_var, likelihood_gp_mean, likelihood_gp_var

         


import numpy as np
from scipy.special import loggamma
from scipy.linalg import cho_solve
from .functions import fmvn_mu

class Poisson:
    """Class to implement Poisson likelihood. It (and all likelihoods below) can only be added as the final
       layer of the DGP+likelihood model.

    Args:
        input_dim (ndarray, optional): a numpy 1d-array that contains the indices of GPs in the last 
            layer whose outputs feed into the likelihood node. When set to `None`, all outputs from GPs of 
            last layer feed into the likelihood node. Defaults to `None`.

    Attributes:
        type (str): identifies that the node is a likelihood node;
        input (ndarray): a numpy 2d-array (each row as a data point and each column as a likelihood parameter from the
            DGP part) that contains the input data (according to the argument **input_dim**) to the likelihood node. The value of 
            this attribute is assigned during the initialisation of :class:`.dgp` class. 
        output (ndarray): a numpy 2d-array with only one column that contains the output data to the likelihood node.
            The value of this attribute is assigned during the initialisation of :class:`.dgp` class.
        exact_post_idx (ndarray): a numpy 1d-array that indicates the indices of the likelihood parameters that allow closed-form
            conditional posterior distributions. Defaults to `None`.
        rep (ndarray): a numpy 1d-array used to re-construct repetitions in the data according to the repetitions in the global input,
            i.e., rep is assigned during the initialisation of :class:`.dgp` class if one input position has multiple outputs. Otherwise, it is
            `None`. Defaults to `None`. 
    """
    def __init__(self, input_dim=None):
        self.type='likelihood'
        self.name='Poisson'
        self.input=None
        self.output=None
        self.input_dim=input_dim
        self.exact_post_idx=None
        self.rep=None

    def llik(self):
        """The log-likelihood function of Poisson distribution.

        Returns:
            ndarray: a numpy 1d-array of log-likelihood.
        """
        llik=self.output*self.input-np.exp(self.input)-loggamma(self.output+1)
        llik=np.sum(llik)
        return llik
    
    @staticmethod
    def pllik(y,f):
        """The predicted log-likelihood function of Poisson distribution.

        Args:
            y (ndarray): a numpy 3d-array of output data with shape ``(N,1,1)``, where *N* is the number of output data points.
            f (ndarray): a numpy 3d-array of sample points with shape ``(N,S,Q)``, where *S* is the number of sample points and 
                *Q* is the number of parameters in the distribution (e.g., *Q* = `1` for Poisson distribution).

        Returns:
            ndarray: a numpy 3d-array of log-likelihood for given **f**.
        """
        pllik=y*f-np.exp(f)-loggamma(y+1)
        return pllik

    @staticmethod    
    def prediction(m,v):
        """Compute mean and variance of the DGP+Poisson model given the predictive
           mean and variance of DGP model for Poisson parameter.
        
        Args:
            m (ndarray): a numpy 2d-array of predictive mean from the DGP model for the Poisson parameter.
            v (ndarray): a numpy 2d-array of predictive variance from the DGP model for the Poisson parameter.

        Returns:
            tuple: a tuple of two 1d-arrays giving the means and variances at the testing input data positions (that are 
            represented by predictive means and variances).
        """
        y_mean=np.exp(m+v/2)
        y_var=np.exp(m+v/2)+(np.exp(v)-1)*np.exp(2*m+v)
        return y_mean.flatten(),y_var.flatten()
    
    def sampling(self,f_sample):
        """Generate samples of DGP+Poisson model given samples of DGP model for the Poisson parameter.
        
        Args:
            f_sample (ndarray): a numpy 2d-array (with one column) of samples from the DGP model for the Poisson parameter.
           
        Returns:
            tuple: a tuple of one 1d-arrays giving samples at the testing input data positions.
        """
        y_sample=np.random.poisson(np.exp(f_sample))
        return y_sample.flatten()

class Hetero:
    def __init__(self, input_dim=None):
        self.type='likelihood'
        self.name='Hetero'
        self.input=None
        self.output=None
        self.input_dim=input_dim
        self.exact_post_idx=np.array([0])
        self.rep=None

    def llik(self):
        mu,var=self.input[:,0],np.exp(self.input[:,1])
        llik=-0.5*(np.log(2*np.pi*var)+((self.output).flatten()-mu)**2/var)
        llik=np.sum(llik)
        return llik

    @staticmethod
    def pllik(y,f):
        mu,var=f[:,:,[0]],np.exp(f[:,:,[1]])
        pllik=-0.5*(np.log(2*np.pi*var)+(y-mu)**2/var)
        return pllik

    @staticmethod    
    def prediction(m,v):
        y_mean=m[:,0]
        y_var=np.exp(m[:,1]+v[:,1]/2)+v[:,0]
        return y_mean.flatten(),y_var.flatten()
    
    @staticmethod
    def sampling(f_sample):
        y_sample=np.random.normal(f_sample[:,0],np.sqrt(np.exp(f_sample[:,1])))
        return y_sample.flatten()

    def posterior(self,idx,v):
        """Sampling from the conditional posterior distribution of the mean in heteroskedastic Gaussian likelihood.
        """
        if idx==0:
            if self.rep is None:
                Gamma=np.diag(np.exp(self.input[:,1]))
                #y=(self.output).flatten()
                mu,cov=self.post_het1(v,Gamma,self.output)
            else:
                Gamma=np.diag(np.exp(self.input[:,1]))
                #y_mask=self.output[:,0]
                mask_f=self.rep
                v_mask=v[mask_f,:]
                V_mask=v[mask_f,:][:,mask_f]
                #mu,cov=self.post_het2(v,Gamma,v_mask,V_mask,y_mask)
                mu,cov=self.post_het2(v,Gamma,v_mask,V_mask,self.output)
            #f_mu=np.random.default_rng().multivariate_normal(mean=mu,cov=cov,check_valid='ignore')
            f_mu=fmvn_mu(mu,cov)
            return f_mu
    
    @staticmethod
    def post_het1(v,Gamma,y_mask):
        """Calculate the conditional posterior mean and covariance of the mean 
           of the heteroskedastic Gaussian likelihood when there are no repetitions
           in the training data.
        """
        L=np.linalg.cholesky(Gamma+v)
        #mu=np.sum(v*cho_solve((L, True), y_mask, check_finite=False),axis=1)
        mu=(v@cho_solve((L, True), y_mask, check_finite=False)).flatten()
        cov=v@cho_solve((L, True), Gamma, check_finite=False)
        return mu, cov

    @staticmethod
    def post_het2(v,Gamma,v_mask,V_mask,y_mask):
        """Calculate the conditional posterior mean and covariance of the mean 
           of the heteroskedastic Gaussian likelihood when there are repetitions
           in the training data.
        """
        L=np.linalg.cholesky(Gamma+V_mask)
        #mu=np.sum(v_mask.T*cho_solve((L, True), y_mask, check_finite=False),axis=1)
        mu=(v_mask.T@cho_solve((L, True), y_mask, check_finite=False)).flatten()
        cov=v-v_mask.T@cho_solve((L, True), v_mask, check_finite=False)
        return mu, cov    

class NegBin:
    def __init__(self, input_dim=None):
        self.type='likelihood'
        self.name='NegBin'
        self.input=None
        self.output=None
        self.input_dim=input_dim
        self.exact_post_idx=None
        self.rep=None
    
    def llik(self):
        y,mu,sigma=(self.output).flatten(),np.exp(self.input[:,0]),np.exp(self.input[:,1])
        llik=loggamma(y+1/sigma)-loggamma(1/sigma)-loggamma(y+1)+y*np.log(sigma*mu)-(y+1/sigma)*np.log(1+sigma*mu)
        llik=np.sum(llik)
        return llik

    @staticmethod
    def pllik(y,f):
        mu,sigma=np.exp(f[:,:,[0]]),np.exp(f[:,:,[1]])
        pllik=loggamma(y+1/sigma)-loggamma(1/sigma)-loggamma(y+1)+y*np.log(sigma*mu)-(y+1/sigma)*np.log(1+sigma*mu)
        return pllik
    
    @staticmethod
    def prediction(m,v):
        y_mean=np.exp(m[:,0]+v[:,0]/2)
        y_var=np.exp(2*m[:,0]+v[:,0])*(np.exp(v[:,0])-1)+np.exp(m[:,0]+v[:,0]/2)+np.exp(m[:,1]+v[:,1]/2)*np.exp(2*m[:,0]+2*v[:,0])
        return y_mean.flatten(),y_var.flatten()
    
    @staticmethod
    def sampling(f_sample):
        p, k=1/(1+np.exp(f_sample[:,0]+f_sample[:,1])), np.exp(-f_sample[:,1])
        y_sample=np.random.negative_binomial(k,p)
        return y_sample.flatten()


        


import multiprocess.context as ctx
import platform
import numpy as np
from pathos.multiprocessing import ProcessingPool as Pool
import psutil   
from .imputation import imputer
import copy
from .functions import ghdiag

class emulator:
    """Class to make predictions from the trained DGP or DGP+likelihood model.

    Args:
        all_layer (list): a list that contains the trained DGP model produced by the method :meth:`.estimate`
            of the :class:`.dgp` class. 
        N (int, optional): the number of imputation to produce the predictions. Increase the value to account for
            more imputation uncertainties. Defaults to `50`.
        nb_parallel (bool, optional): whether to use *Numba*'s multi-threading to accelerate the predictions. Defaults to `False`.
    """
    def __init__(self, all_layer, N=50, nb_parallel=False):
        self.all_layer=all_layer
        self.n_layer=len(all_layer)
        self.imp=imputer(self.all_layer)
        (self.imp).sample(burnin=50)
        self.all_layer_set=[]
        for _ in range(N):
            (self.imp).sample()
            (self.imp).key_stats()
            (self.all_layer_set).append(copy.deepcopy(self.all_layer))
        self.nb_parallel=nb_parallel
        #if len(self.all_layer[0][0].input)>=500 and self.nb_parallel==False:
        #    print('Your training data size is greater than %i, you might want to set "nb_parallel=True" to accelerate the prediction.' % (500))
    
    def set_nb_parallel(self,nb_parallel):
        """Set **self.nb_parallel** to the bool value given by **nb_parallel**. This method is useful to change **self.nb_parallel**
            when the emulator class has already been built.
        """
        self.nb_parallel=nb_parallel

    def ppredict(self,x,method='mean_var',full_layer=False,sample_size=50,chunk_num=None,core_num=None):
        """Implement parallel predictions from the trained DGP model.

        Args:
            x, method, full_layer, sample_size: see descriptions of the method :meth:`.emulator.predict`.
            chunk_num (int, optional): the number of chunks that the testing input array **x** will be divided into. 
                Defaults to `None`. If not specified, the number of chunks is set to **core_num**. 
            core_num (int, optional): the number of cores/workers to be used. Defaults to `None`. If not specified, 
                the number of cores is set to ``(max physical cores available - 1)``.

        Returns:
            Same as the method :meth:`.emulator.predict`.
        """
        if platform.system()=='Darwin':
            ctx._force_start_method('forkserver')
        if core_num==None:
            core_num=psutil.cpu_count(logical = False)-1
        if chunk_num==None:
            chunk_num=core_num
        if chunk_num<core_num:
            core_num=chunk_num
        f=lambda x: self.predict(*x) 
        z=np.array_split(x,chunk_num)
        with Pool(core_num) as pool:
            #pool.restart()
            res = pool.map(f, [[x, method, full_layer, sample_size] for x in z])
            pool.close()
            pool.join()
            pool.clear()
        if method == 'mean_var':
            if full_layer:
                combined_res=[]
                for layer in zip(*res):
                    combined_res.append(list(np.concatenate(workers) for workers in zip(*list(layer))))
                return tuple(combined_res)
            else:
                return tuple(np.concatenate(worker) for worker in zip(*res))
        elif method == 'sampling':
            if full_layer:
                combined_res=[]
                for layer in zip(*res):
                    combined_res.append(list(np.concatenate(workers) for workers in zip(*list(layer))))
                return combined_res
            else:
                return list(np.concatenate(worker) for worker in zip(*res))

    def predict(self,x,method='mean_var',full_layer=False,sample_size=50):
        """Implement predictions from the trained DGP model.

        Args:
            x (ndarray): a numpy 2d-array where each row is an input testing data point and 
                each column is an input dimension.
            method (str, optional): the prediction approach: mean-variance (`mean_var`) or sampling 
                (`sampling`) approach. Defaults to `mean_var`.
            full_layer (bool, optional): whether to output the predictions of all layers. Defaults to `False`.
            sample_size (int, optional): the number of samples to draw for each given imputation if **method** = '`sampling`'.
                 Defaults to `50`.
            
        Returns:
            tuple_or_list: if the argument **method** = '`mean_var`', a tuple is returned:
                1. If **full_layer** = `False`, the tuple contains two numpy 2d-arrays, one for the predictive means 
                   and another for the predictive variances. Each array has its rows corresponding to testing 
                   positions and columns corresponding to DGP output dimensions (i.e., GP/likelihood nodes in the final layer);
                2. If **full_layer** = `True`, the tuple contains two lists, one for the predictive means 
                   and another for the predictive variances. Each list contains *L* (i.e., the number of layers) 
                   numpy 2d-arrays. Each array has its rows corresponding to testing positions and columns 
                   corresponding to output dimensions (i.e., GP nodes from the associated layer and in case of the final layer, 
                   it may be the number of the likelihood nodes).
            if the argument **method** = '`sampling`', a list is returned:
                1. If **full_layer** = `False`, the list contains *D* (i.e., the number of GP/likelihood nodes in the final layer) numpy 
                   2d-arrays. Each array has its rows corresponding to testing positions and columns corresponding to samples of
                   size: **N** * **sample_size**;
                2. If **full_layer** = `True`, the list contains *L* (i.e., the number of layers) sub-lists. Each sub-list 
                   represents the samples draw from the GPs/likelihoods in the corresponding layers, and contains 
                   *D* (i.e., the number of GP nodes in the corresponding layer or likelihood nodes in the final layer) 
                   numpy 2d-arrays. Each array gives samples of the output from one of *D* GPs/likelihoods at the 
                   testing positions, and has its rows corresponding to testing positions and columns corresponding to samples
                   of size: **N** * **sample_size**.
        """
        if x.ndim==1:
            raise Exception('The testing input has to be a numpy 2d-array')
        M=len(x)
        if method=='mean_var':
            sample_size=1
        #start predictions
        mean_pred=[]
        variance_pred=[]
        likelihood_mean=[]
        likelihood_variance=[]
        for s in range(len(self.all_layer_set)):
            overall_global_test_input=x
            one_imputed_all_layer=self.all_layer_set[s]
            if full_layer:
                mean_pred_oneN=[]
                variance_pred_oneN=[]
            for l in range(self.n_layer):
                layer=one_imputed_all_layer[l]
                n_kerenl=len(layer)
                if l==self.n_layer-1:
                    likelihood_gp_mean=np.empty((M,n_kerenl))
                    likelihood_gp_var=np.empty((M,n_kerenl))
                else:
                    overall_test_output_mean=np.empty((M,n_kerenl))
                    overall_test_output_var=np.empty((M,n_kerenl))
                if l==0:
                    for k in range(n_kerenl):
                        kernel=layer[k]
                        if kernel.connect is not None:
                            z_k_in=overall_global_test_input[:,kernel.connect]
                        else:
                            z_k_in=None
                        m_k,v_k=kernel.gp_prediction(x=overall_global_test_input[:,kernel.input_dim],z=z_k_in)
                        overall_test_output_mean[:,k],overall_test_output_var[:,k]=m_k,v_k
                    overall_test_input_mean,overall_test_input_var=overall_test_output_mean,overall_test_output_var
                    if full_layer:
                        mean_pred_oneN.append(overall_test_input_mean)
                        variance_pred_oneN.append(overall_test_input_var)
                elif l==self.n_layer-1:
                    for k in range(n_kerenl):
                        kernel=layer[k]
                        m_k_in,v_k_in=overall_test_input_mean[:,kernel.input_dim],overall_test_input_var[:,kernel.input_dim]
                        if kernel.type=='gp':
                            if kernel.connect is not None:
                                z_k_in=overall_global_test_input[:,kernel.connect]
                            else:
                                z_k_in=None
                            m_k,v_k=kernel.linkgp_prediction(m=m_k_in,v=v_k_in,z=z_k_in,nb_parallel=self.nb_parallel)
                            likelihood_gp_mean[:,k],likelihood_gp_var[:,k]=m_k,v_k
                        elif kernel.type=='likelihood':
                            m_k,v_k=kernel.prediction(m=m_k_in,v=v_k_in)
                            likelihood_gp_mean[:,k],likelihood_gp_var[:,k]=m_k,v_k
                else:
                    for k in range(n_kerenl):
                        kernel=layer[k]
                        m_k_in,v_k_in=overall_test_input_mean[:,kernel.input_dim],overall_test_input_var[:,kernel.input_dim]
                        if kernel.connect is not None:
                            z_k_in=overall_global_test_input[:,kernel.connect]
                        else:
                            z_k_in=None
                        m_k,v_k=kernel.linkgp_prediction(m=m_k_in,v=v_k_in,z=z_k_in,nb_parallel=self.nb_parallel)
                        overall_test_output_mean[:,k],overall_test_output_var[:,k]=m_k,v_k
                    overall_test_input_mean,overall_test_input_var=overall_test_output_mean,overall_test_output_var
                    if full_layer:
                        mean_pred_oneN.append(overall_test_input_mean)
                        variance_pred_oneN.append(overall_test_input_var)
            for _ in range(sample_size):
                if full_layer:
                    mean_pred.append(mean_pred_oneN)
                    variance_pred.append(variance_pred_oneN)
                else:
                    mean_pred.append(overall_test_input_mean)
                    variance_pred.append(overall_test_input_var)
                likelihood_mean.append(likelihood_gp_mean)
                likelihood_variance.append(likelihood_gp_var)
        if method=='sampling':
            if full_layer:
                mu_layerwise=[list(mean_n) for mean_n in zip(*mean_pred)]
                var_layerwise=[list(var_n) for var_n in zip(*variance_pred)]
                samples=[]
                for l in range(self.n_layer):
                    samples_layerwise=[]
                    if l==self.n_layer-1:
                        for mu_likelihood, sigma2_likelihood, dgp_sample in zip(likelihood_mean, likelihood_variance, samples_layer_before_likelihood):
                            realisation=np.empty_like(mu_likelihood)
                            for count, kernel in enumerate(self.all_layer[-1]):
                                if kernel.type=='gp':
                                    realisation[:,count]=np.random.normal(mu_likelihood[:,count],np.sqrt(sigma2_likelihood[:,count]))
                                elif kernel.type=='likelihood':
                                    realisation[:,count]=kernel.sampling(dgp_sample[:,kernel.input_dim])
                            samples_layerwise.append(realisation)
                    else:
                        for mu, sigma2 in zip(mu_layerwise[l], var_layerwise[l]):
                            realisation=np.random.normal(mu,np.sqrt(sigma2))
                            samples_layerwise.append(realisation)
                        if l==self.n_layer-2:
                            samples_layer_before_likelihood=samples_layerwise
                    samples_layerwise=np.asarray(samples_layerwise).transpose(2,1,0)
                    samples.append(list(samples_layerwise))
            else:
                samples=[]
                for mu_dgp, sigma2_dgp, mu_likelihood, sigma2_likelihood  in zip(mean_pred, variance_pred, likelihood_mean, likelihood_variance):
                    realisation=np.empty_like(mu_likelihood)
                    for count, kernel in enumerate(self.all_layer[-1]):
                        if kernel.type=='gp':
                            realisation[:,count]=np.random.normal(mu_likelihood[:,count],np.sqrt(sigma2_likelihood[:,count]))
                        elif kernel.type=='likelihood':
                            dgp_sample=np.random.normal(mu_dgp,np.sqrt(sigma2_dgp))
                            realisation[:,count]=kernel.sampling(dgp_sample[:,kernel.input_dim])
                    samples.append(realisation)
                samples=list(np.asarray(samples).transpose(2,1,0))
            return samples
        elif method=='mean_var':
            if full_layer:
                mu_layerwise=[list(mean_n) for mean_n in zip(*mean_pred)]
                var_layerwise=[list(var_n) for var_n in zip(*variance_pred)]
                mu=[np.mean(mu_l,axis=0) for mu_l in mu_layerwise]
                mu2_mean=[np.mean(np.square(mu_l),axis=0) for mu_l in mu_layerwise]
                var_mean=[np.mean(var_l,axis=0) for var_l in var_layerwise]
                sigma2=[i+j-k**2 for i,j,k in zip(mu2_mean,var_mean,mu)]
                mu.append(np.mean(likelihood_mean,axis=0))
                sigma2.append(np.mean((np.square(likelihood_mean)+likelihood_variance),axis=0)-np.mean(likelihood_mean,axis=0)**2)
            else:
                mu=np.mean(likelihood_mean,axis=0)
                sigma2=np.mean((np.square(likelihood_mean)+likelihood_variance),axis=0)-mu**2
            return mu, sigma2

    def nllik(self,x,y):
        """Compute the negative predicted log-likelihood from the trained DGP model with likelihood layer.

        Args:
            x (ndarray): a numpy 2d-array where each row is an input testing data point and 
                each column is an input dimension.
            y (ndarray): a numpy 2d-array where each row is a scalar-valued testing output data point.

        Returns:
            tuple: a tuple of two 1d-arrays. The first one is the average negative predicted log-likelihood across
            all testing data points. The second ones is the negative predicted log-likelihood for each testing data point.
        """
        if len(self.all_layer[-1])!=1:
            raise Exception('The method is only applicable to DGP with the final layer formed by only ONE node, which must be a likelihood node.')
        else:
            if self.all_layer[-1][0].type!='likelihood':
                raise Exception('The method is only applicable to DGP with the final layer formed by only ONE likelihood node, which must be a likelihood node.')
        M=len(x)
        #start predictions
        predicted_lik=[]
        for s in range(len(self.all_layer_set)):
            overall_global_test_input=x
            one_imputed_all_layer=self.all_layer_set[s]
            for l in range(self.n_layer-1):
                layer=one_imputed_all_layer[l]
                n_kerenl=len(layer)
                overall_test_output_mean=np.empty((M,n_kerenl))
                overall_test_output_var=np.empty((M,n_kerenl))
                if l==0:
                    for k in range(n_kerenl):
                        kernel=layer[k]
                        if kernel.connect is not None:
                            z_k_in=overall_global_test_input[:,kernel.connect]
                        else:
                            z_k_in=None
                        m_k,v_k=kernel.gp_prediction(x=overall_global_test_input[:,kernel.input_dim],z=z_k_in)
                        overall_test_output_mean[:,k],overall_test_output_var[:,k]=m_k,v_k
                    overall_test_input_mean,overall_test_input_var=overall_test_output_mean,overall_test_output_var
                else:
                    for k in range(n_kerenl):
                        kernel=layer[k]
                        m_k_in,v_k_in=overall_test_input_mean[:,kernel.input_dim],overall_test_input_var[:,kernel.input_dim]
                        if kernel.connect is not None:
                            z_k_in=overall_global_test_input[:,kernel.connect]
                        else:
                            z_k_in=None
                        m_k,v_k=kernel.linkgp_prediction(m=m_k_in,v=v_k_in,z=z_k_in,nb_parallel=self.nb_parallel)
                        overall_test_output_mean[:,k],overall_test_output_var[:,k]=m_k,v_k
                    overall_test_input_mean,overall_test_input_var=overall_test_output_mean,overall_test_output_var
            predicted_lik.append(ghdiag(one_imputed_all_layer[-1][0].pllik,overall_test_input_mean,overall_test_input_var,y))
        nllik=-np.log(np.mean(predicted_lik,axis=0)).flatten()
        average_nllik=np.mean(nllik)
        return average_nllik, nllik

        
      


import copy
import numpy as np
from .functions import pdist_matern_coef
from scipy.spatial.distance import pdist, squareform

class path:
    #main algorithm
    def __init__(self, X, all_layer):
        self.X=X
        self.n_layer=len(all_layer)
        self.all_layer=copy.deepcopy(all_layer)
        for l in range(self.n_layer):
            layer=self.all_layer[l]
            num_kernel=len(layer)
            for k in range(num_kernel):
                kernel=layer[k]
                if np.any(kernel.connect!=None):
                    kernel.global_input=copy.deepcopy(self.X[:,kernel.connect])

    def generate(self,N):
        d=len(self.all_layer[-1])
        m=len(self.X)
        path_record=np.empty((N,m,d))
        for i in range(N):
            x=self.X
            for l in range(self.n_layer):
                layer=self.all_layer[l]
                num_kernel=len(layer)
                out=np.empty((m,num_kernel))
                for k in range(num_kernel):
                    kernel=layer[k]
                    if np.any(kernel.input_dim!=None):
                        In=x[:,kernel.input_dim]
                    else:
                        In=x
                    if kernel.connect is not None:
                        In=np.concatenate((In,kernel.global_input),1)
                    cov=(self.k_matrix(In,kernel.length,kernel.name)+kernel.nugget*np.identity(m))*kernel.scale
                    L=np.linalg.cholesky(cov)
                    randn=np.random.normal(size=[m,1])
                    out[:,k]=(L@randn).flatten()
                x=out
            path_record[i,]=x
        return path_record.transpose(2,0,1)
    
    @staticmethod
    def k_matrix(X, length, name):
        X_l=X/length
        if name=='sexp':
            dists = pdist(X_l, metric="sqeuclidean")
            K = squareform(np.exp(-dists))
        elif name=='matern2.5':
            K=np.exp(-np.sqrt(5)*pdist(X_l, metric="minkowski",p=1))
            K*=pdist_matern_coef(X_l)
            K=squareform(K)
        np.fill_diagonal(K, 1)
        return K


from numba import jit, vectorize, float64, prange, config
import numpy as np
from math import erf, exp, sqrt, pi
from numpy.random import randn
import itertools
config.THREADING_LAYER = 'workqueue'
#######functions for optim#########
@jit(nopython=True,cache=True)
def pdist_matern_coef(X):
    n = X.shape[0]
    out_size = np.int((n * (n - 1)) / 2)
    dm = np.empty(out_size)
    k = 0
    for i in range(n - 1):
        for j in range(i + 1, n):
            dm[k] = matern_coef(X[i], X[j])
            k += 1
    return dm

@jit(nopython=True,cache=True)
def matern_coef(v,u):
    dist = 1
    for r in range(len(v)):
        disi = np.abs(v[r] - u[r])
        dist *= 1+np.sqrt(5)*disi+(5/3)*disi**2
    return dist

@jit(nopython=True,cache=True)
def fod_exp(X,K):
    n, D = X.shape
    dm = np.zeros((D,n,n))
    for d in range(D):
        for i in range(n - 1):
            for j in range(i + 1, n):
                temp = 2*(X[i,d]-X[j,d])**2*K[i,j]
                dm[d,i,j], dm[d,j,i] = temp, temp
    return dm

@jit(nopython=True,cache=True)
def pdist_matern_one(X):
    n = X.shape[0]
    dm = np.zeros((n,n))
    dm1 = np.zeros((1,n,n))
    for i in range(n - 1):
        for j in range(i + 1, n):
            temp1,temp2 = matern_one(X[i], X[j])
            dm[i,j], dm[j,i] = temp1, temp1 
            dm1[:,i,j], dm1[:,j,i] = temp2, temp2
    return dm, dm1

@jit(nopython=True,cache=True)
def pdist_matern_multi(X):
    n, D = X.shape
    dm = np.zeros((n,n))
    dm2 = np.zeros((D,n,n))
    for i in range(n - 1):
        for j in range(i + 1, n):
            temp1,temp2=matern_multi(X[i], X[j])
            dm[i,j], dm[j,i] = temp1, temp1 
            dm2[:,i,j], dm2[:,j,i] = temp2, temp2
    return dm, dm2

@jit(nopython=True,cache=True)
def matern_one(v,u):
    dist = 1
    dist1 = 0
    for d in range(len(v)):
        disi = np.abs(v[d] - u[d])
        coefi = 1+np.sqrt(5)*disi+(5/3)*disi**2
        dist *= coefi
        coefi1=(5/3)*(disi**2)*(1+np.sqrt(5)*disi)/(1+np.sqrt(5)*disi+5/3*disi**2)
        dist1 += coefi1
    return dist, dist1

@jit(nopython=True,cache=True)
def matern_multi(v,u):
    dist = 1
    dist1=np.zeros(len(v)) 
    for d in range(len(v)):
        disi = np.abs(v[d] - u[d])
        coefi = 1+np.sqrt(5)*disi+(5/3)*disi**2
        dist *= coefi
        coefi1=(5/3)*(disi**2)*(1+np.sqrt(5)*disi)/(1+np.sqrt(5)*disi+5/3*disi**2)
        dist1[d] = coefi1
    return dist, dist1

######functions for imputer########
@jit(nopython=True,cache=True)
def fmvn_mu(mu,cov):
    """Generate multivariate Gaussian random samples with means.
    """
    d=len(cov)
    sn=randn(d,1)
    L=np.linalg.cholesky(cov)
    samp=(L@sn).flatten()+mu
    return samp

@jit(nopython=True,cache=True)
def fmvn(cov):
    """Generate multivariate Gaussian random samples without means.
    """
    d=len(cov)
    sn=randn(d,1)
    L=np.linalg.cholesky(cov)
    samp=(L@sn).flatten()
    return samp

#@jit(nopython=True,cache=True)
#def fmvn_mu(mu,cov):
#    """Generate multivariate Gaussian random samples with means.
#    """
#    d=len(cov)
#    sn=randn(d,1)
#    U, s, _ = np.linalg.svd(cov)
#    samp=((U*np.sqrt(s))@sn).flatten() + mu
#    return samp

#@jit(nopython=True,cache=True)
#def fmvn(cov):
#    """Generate multivariate Gaussian random samples without means.
#    """
#    d=len(cov)
#    sn=randn(d,1)
#    U, s, _ = np.linalg.svd(cov)
#    samp=((U*np.sqrt(s))@sn).flatten()
#    return samp

@jit(nopython=True,cache=True)
def update_f(f,nu,theta):
    """Update ESS proposal samples.
    """
    fp=f*np.cos(theta) + nu*np.sin(theta)
    return fp

@jit(nopython=True,cache=True)
def Z_fct(X,W,b,length,M):
    W=W/length
    Z=np.dot(X,W.T)+b
    return np.sqrt(2/M)*np.cos(Z)

######Gauss-Hermite quadrature######
def ghdiag(fct,mu,var,y):
    x, w = np.polynomial.hermite.hermgauss(10)
    N = np.shape(mu)[1]
    const = np.pi**(-0.5*N)
    xn = np.array(list(itertools.product(*(x,)*N)))
    wn = np.prod(np.array(list(itertools.product(*(w,)*N))), 1)[:, None]
    fn = sqrt(2.0)*(np.sqrt(var[:,None])*xn) + mu[:,None]
    llik=fct(y[:,None],fn)
    return np.sum(np.exp(np.log((wn * const)[None,:]) + llik), axis=1)

######functions for predictions########
#@jit(nopython=True,cache=True,fastmath=True)
def gp(x,z,w1,global_w1,Rinv,Rinv_y,scale,length,nugget,name):
    """Make GP predictions.
    """
    if z is not None:
        x=np.concatenate((x, z),1)
        w1=np.concatenate((w1, global_w1),1)
    r=k_one_vec(w1,x,length,name)
    Rinv_r=np.dot(Rinv,r)
    r_Rinv_r=np.sum(r*Rinv_r,axis=0)
    v=np.abs(scale*(1+nugget-r_Rinv_r))
    #m=np.sum(r*Rinv_y,axis=0) 
    m=(r.T@Rinv_y).flatten()
    return m, v

#@jit(nopython=True,cache=True,fastmath=True)
def link_gp(m,v,z,w1,global_w1,Rinv,Rinv_y,scale,length,nugget,name,nb_parallel):
    """Make linked GP predictions.
    """
    M=len(m)
    m_new=np.empty((M,1))
    v_new=np.empty((M,1))
    if z is not None:
        Dw=np.shape(w1)[1]
        Dz=np.shape(z)[1]
        if len(length)==1:
            length=length*np.ones(Dw+Dz)
    else:
        Dw=np.shape(w1)[1]
        if len(length)==1:
            length=length*np.ones(Dw)
    for i in range(M):
        mi=m[i,:]
        vi=v[i,:]
        if z is not None:
            zi=np.expand_dims(z[i,:],axis=0)
            if nb_parallel:
                I,J=IJ_parallel(w1,mi,vi,length[:-Dz],name)
            else:
                I,J=IJ(w1,mi,vi,length[:-Dz],name)
            Iz=k_one_vec(global_w1,zi,length[-Dz::],name)
            #Iz_T=np.expand_dims(Iz.flatten(),axis=0)
            Jz=np.dot(Iz,Iz.T)
            I,J=I*Iz,J*Jz
        else:
            I,J=IJ(w1,mi,vi,length,name)
        tr_RinvJ=np.sum(Rinv*J)
        #IRinv_y=np.sum(I*Rinv_y)
        IRinv_y=I.T@Rinv_y
        m_new[i]=IRinv_y
        #v_new[i]=np.abs(np.sum(np.sum(Rinv_y*J,axis=0)*Rinv_y.flatten())-IRinv_y**2+scale*(1+nugget-tr_RinvJ))
        v_new[i]=np.abs(Rinv_y.T@J@Rinv_y-IRinv_y**2+scale*(1+nugget-tr_RinvJ))
    return m_new.flatten(),v_new.flatten()

@jit(nopython=True,cache=True,fastmath=True)
def k_one_vec(X,z,length,name):
    """Compute cross-correlation matrix between the testing and training input data.
    """
    if name=='sexp':
        X_l=X/length
        z_l=z/length
        L_X=np.expand_dims(np.sum(X_l**2,axis=1),axis=1)
        L_z=np.expand_dims(np.sum(z_l**2,axis=1,),axis=1)
        dis2=L_X-2*np.dot(X_l,z_l.T)+L_z.T
        k=np.exp(-dis2)
    elif name=='matern2.5':
        n=np.shape(X)[0]
        d=np.shape(X)[1]
        m=len(z)
        X_l=np.expand_dims((X/length).T,axis=2)
        z_l=np.expand_dims((z/length).T,axis=2)
        k1=np.ones((n,m))
        k2=np.zeros((n,m))
        for i in range(d):
            dis=np.abs(X_l[i]-z_l[i].T)
            k1*=(1+sqrt(5)*dis+5/3*dis**2)
            k2+=dis
        k2=np.exp(-sqrt(5)*k2)
        k=k1*k2
    return k

@jit(nopython=True,cache=True,parallel=True)
def IJ_parallel(X,z_m,z_v,length,name):
    """Compute I and J involved in linked GP predictions.
    """
    n = X.shape[0]
    if name=='sexp':
        X_z=X-z_m
        I=np.ones((n,1))
        J=np.ones((n,n))
        for i in prange(n):
            I[i]=I_sexp_parallel(z_v,length,X_z[i])
            for j in range(i+1):
                temp = J_sexp_parallel(z_v,length,X_z[i],X_z[j])
                if i==j:
                    J[i,j]=temp
                else:
                    J[i,j], J[j,i]=temp, temp
    elif name=='matern2.5':
        zX=z_m-X
        muA, muB=zX-sqrt(5)*z_v/length, zX+sqrt(5)*z_v/length
        I=np.ones((n,1))
        J=np.ones((n,n))
        for i in prange(n):
            I[i]=I_matern_parallel(z_v,length,muA[i],muB[i],zX[i])
            for j in range(i+1):
                temp = J_matern_parallel(z_m, z_v, length, X[i], X[j], zX[i], zX[j])
                if i==j:
                    J[i,j]=temp
                else:
                    J[i,j], J[j,i]=temp, temp 
    return I,J

@jit(nopython=True,cache=True)
def I_sexp_parallel(z_v,length,X_zi):
    Id=1
    for d in range(len(length)):
        Id*=1/np.sqrt(1+2*z_v[d]/length[d]**2)*np.exp(-X_zi[d]**2/(2*z_v[d]+length[d]**2))
    return Id

@jit(nopython=True,cache=True)
def J_sexp_parallel(z_v,length,X_zi,X_zj):
    Jd=1
    for d in range(len(length)):
        dis1, dis2 = (X_zi[d]+X_zj[d])**2, (X_zi[d]-X_zj[d])**2
        Jd*=1/np.sqrt(1+4*z_v[d]/length[d]**2)*np.exp(-dis1/(2*length[d]**2+8*z_v[d])-dis2/(2*length[d]**2))
    return Jd

@jit(nopython=True,cache=True)
def I_matern_parallel(z_v,length,muA,muB,zXi):
    Id=1
    for d in range(len(length)):
        if z_v[d]!=0:
            Id*=np.exp((5*z_v[d]-2*sqrt(5)*length[d]*zXi[d])/(2*length[d]**2))* \
                ((1+sqrt(5)*muA[d]/length[d]+5*(muA[d]**2+z_v[d])/(3*length[d]**2))*0.5*(1+erf(muA[d]/sqrt(2*z_v[d])))+ \
                (sqrt(5)+(5*muA[d])/(3*length[d]))*sqrt(0.5*z_v[d]/pi)/length[d]*np.exp(-0.5*muA[d]**2/z_v[d]))+ \
                np.exp((5*z_v[d]+2*sqrt(5)*length[d]*zXi[d])/(2*length[d]**2))* \
                ((1-sqrt(5)*muB[d]/length[d]+5*(muB[d]**2+z_v[d])/(3*length[d]**2))*0.5*(1+erf(-muB[d]/sqrt(2*z_v[d])))+ \
                (sqrt(5)-(5*muB[d])/(3*length[d]))*sqrt(0.5*z_v[d]/pi)/length[d]*np.exp(-0.5*muB[d]**2/z_v[d]))
        else:
            Id*=(1+sqrt(5)*np.abs(zXi[d])/length[d]+5*zXi[d]**2/(3*length[d]**2))*np.exp(-sqrt(5)*np.abs(zXi[d])/length[d])  
    return Id

@jit(nopython=True,cache=True)
def J_matern_parallel(z_m, z_v, length, Xi, Xj, zXi, zXj):
    J_d=1
    for d in range(len(length)):
        if z_v[d]!=0:
            J_d*=Jd(Xj[d],Xi[d],z_m[d],z_v[d],length[d])
        else:
            Idi=(1+sqrt(5)*np.abs(zXi[d])/length[d]+5*zXi[d]**2/(3*length[d]**2))*np.exp(-sqrt(5)*np.abs(zXi[d])/length[d])
            Idj=(1+sqrt(5)*np.abs(zXj[d])/length[d]+5*zXj[d]**2/(3*length[d]**2))*np.exp(-sqrt(5)*np.abs(zXj[d])/length[d])
            J_d*=(Idi*Idj)
    return J_d

@jit(nopython=True,cache=True,fastmath=True)
def IJ(X,z_m,z_v,length,name):
    """Compute I and J involved in linked GP predictions.
    """
    n,d=X.shape
    if name=='sexp':
        X_z=(X-z_m).T
        I=np.ones((n))
        J=np.ones((n,n))
        for i in range(d):
            I*=1/np.sqrt(1+2*z_v[i]/length[i]**2)*np.exp(-X_z[i]**2/(2*z_v[i]+length[i]**2))
            for k in range(n):
                for l in range(k+1):
                    temp = J_sexp(z_v[i],length[i],X_z[i,k],X_z[i,l])
                    if k==l:
                        J[k,l]*=temp
                    else:
                        J[k,l]*=temp
                        J[l,k]*=temp
        return I.reshape(-1,1),J
    elif name=='matern2.5':
        zX=z_m-X
        muA=(zX-sqrt(5)*z_v/length).T.reshape((d,n,1))
        muB=(zX+sqrt(5)*z_v/length).T.reshape((d,n,1))
        zX=zX.T.reshape((d,n,1))
        X=X.T.reshape((d,n,1))
        I=np.ones((n,1))
        J=np.ones((n,n))
        for i in range(d):    
            if z_v[i]!=0:
                I*=np.exp((5*z_v[i]-2*sqrt(5)*length[i]*zX[i])/(2*length[i]**2))* \
                    ((1+sqrt(5)*muA[i]/length[i]+5*(muA[i]**2+z_v[i])/(3*length[i]**2))*pnorm(muA[i]/sqrt(z_v[i]))+ \
                    (sqrt(5)+(5*muA[i])/(3*length[i]))*sqrt(0.5*z_v[i]/pi)/length[i]*np.exp(-0.5*muA[i]**2/z_v[i]))+ \
                   np.exp((5*z_v[i]+2*sqrt(5)*length[i]*zX[i])/(2*length[i]**2))* \
                    ((1-sqrt(5)*muB[i]/length[i]+5*(muB[i]**2+z_v[i])/(3*length[i]**2))*pnorm(-muB[i]/sqrt(z_v[i]))+ \
                    (sqrt(5)-(5*muB[i])/(3*length[i]))*sqrt(0.5*z_v[i]/pi)/length[i]*np.exp(-0.5*muB[i]**2/z_v[i]))
                for k in range(n):
                    for l in range(k+1):
                        J_lk=Jd(X[i][l][0],X[i][k][0],z_m[i],z_v[i],length[i])
                        if l==k:
                            J[l,k]*=J_lk
                        else:
                            J[l,k]*=J_lk
                            J[k,l]*=J_lk
            else:
                Id=(1+sqrt(5)*np.abs(zX[i])/length[i]+5*zX[i]**2/(3*length[i]**2))*np.exp(-sqrt(5)*np.abs(zX[i])/length[i])
                I*=Id
                J*=np.dot(Id,Id.T)
        return I,J

@jit(nopython=True,cache=True,fastmath=True)
def J_sexp(z_v,length,X_zi,X_zj):
    dis1, dis2 = (X_zi+X_zj)**2, (X_zi-X_zj)**2
    Jd=1/sqrt(1+4*z_v/length**2)*exp(-dis1/(2*length**2+8*z_v)-dis2/(2*length**2))
    return Jd

@vectorize([float64(float64)],nopython=True,cache=True,fastmath=True)
def pnorm(x):
    """Compute standard normal CDF.
    """
    return 0.5*(1+erf(x/sqrt(2)))    

@jit(nopython=True,cache=True,fastmath=True)
def Jd(X1,X2,z_m,z_v,length):
    """Compute J components in different input dimensions for Matern2.5 kernel.
    """
    if X1<X2:
        x1=X1
        x2=X2
    else:
        x1=X2
        x2=X1
    E30=1+(25*x1**2*x2**2-3*sqrt(5)*(3*length**3+5*length*x1*x2)*(x1+x2)+15*length**2*(x1**2+x2**2+3*x1*x2))/(9*length**4)
    E31=(18*sqrt(5)*length**3+15*sqrt(5)*length*(x1**2+x2**2)-(75*length**2+50*x1*x2)*(x1+x2)+60*sqrt(5)*length*x1*x2)/(9*length**4)
    E32=5*(5*x1**2+5*x2**2+15*length**2-9*sqrt(5)*length*(x1+x2)+20*x1*x2)/(9*length**4)
    E33=10*(3*sqrt(5)*length-5*x1-5*x2)/(9*length**4)
    E34=25/(9*length**4)
    muC=z_m-2*sqrt(5)*z_v/length
    E3A31=E30+muC*E31+(muC**2+z_v)*E32+(muC**3+3*z_v*muC)*E33+(muC**4+6*z_v*muC**2+3*z_v**2)*E34
    E3A32=E31+(muC+x2)*E32+(muC**2+2*z_v+x2**2+muC*x2)*E33+(muC**3+x2**3+x2*muC**2+muC*x2**2+3*z_v*x2+5*z_v*muC)*E34
    P1=exp((10*z_v+sqrt(5)*length*(x1+x2-2*z_m))/length**2)*(0.5*E3A31*(1+erf((muC-x2)/sqrt(2*z_v)))+\
        E3A32*sqrt(0.5*z_v/pi)*exp(-0.5*(x2-muC)**2/z_v))
    
    E40=1+(25*x1**2*x2**2+3*sqrt(5)*(3*length**3-5*length*x1*x2)*(x2-x1)+15*length**2*(x1**2+x2**2-3*x1*x2))/(9*length**4)
    E41=5*(3*sqrt(5)*length*(x2**2-x1**2)+3*length**2*(x1+x2)-10*x1*x2*(x1+x2))/(9*length**4)
    E42=5*(5*x1**2+5*x2**2-3*length**2-3*sqrt(5)*length*(x2-x1)+20*x1*x2)/(9*length**4)
    E43=-50*(X1+X2)/(9*length**4)
    E44=25/(9*length**4)
    E4A41=E40+z_m*E41+(z_m**2+z_v)*E42+(z_m**3+3*z_v*z_m)*E43+(z_m**4+6*z_v*z_m**2+3*z_v**2)*E44
    E4A42=E41+(z_m+x1)*E42+(z_m**2+2*z_v+x1**2+z_m*x1)*E43+(z_m**3+x1**3+x1*z_m**2+z_m*x1**2+3*z_v*x1+5*z_v*z_m)*E44
    E4A43=E41+(z_m+x2)*E42+(z_m**2+2*z_v+x2**2+z_m*x2)*E43+(z_m**3+x2**3+x2*z_m**2+z_m*x2**2+3*z_v*x2+5*z_v*z_m)*E44
    P2=exp(-sqrt(5)*(x2-x1)/length)*(0.5*E4A41*(erf((x2-z_m)/sqrt(2*z_v))-erf((x1-z_m)/sqrt(2*z_v)))+\
        E4A42*sqrt(0.5*z_v/pi)*exp(-0.5*(x1-z_m)**2/z_v)-E4A43*sqrt(0.5*z_v/pi)*exp(-0.5*(x2-z_m)**2/z_v))

    E50=1+(25*x1**2*x2**2+3*sqrt(5)*(3*length**3+5*length*x1*x2)*(x1+x2)+15*length**2*(x1**2+x2**2+3*x1*x2))/(9*length**4)
    E51=(18*sqrt(5)*length**3+15*sqrt(5)*length*(x1**2+x2**2)+(75*length**2+50*x1*x2)*(x1+x2)+60*sqrt(5)*length*x1*x2)/(9*length**4)
    E52=5*(5*x1**2+5*x2**2+15*length**2+9*sqrt(5)*length*(x1+x2)+20*x1*x2)/(9*length**4)
    E53=10*(3*sqrt(5)*length+5*x1+5*x2)/(9*length**4)
    E54=25/(9*length**4)
    muD=z_m+2*sqrt(5)*z_v/length
    E5A51=E50-muD*E51+(muD**2+z_v)*E52-(muD**3+3*z_v*muD)*E53+(muD**4+6*z_v*muD**2+3*z_v**2)*E54
    E5A52=E51-(muD+x1)*E52+(muD**2+2*z_v+x1**2+muD*x1)*E53-(muD**3+x1**3+x1*muD**2+muD*x1**2+3*z_v*x1+5*z_v*muD)*E54
    P3=exp((10*z_v-sqrt(5)*length*(x1+x2-2*z_m))/length**2)*(0.5*E5A51*(1+erf((x1-muD)/sqrt(2*z_v)))+\
        E5A52*sqrt(0.5*z_v/pi)*exp(-0.5*(x1-muD)**2/z_v))

    jd=P1+P2+P3
    return jd


import numpy as np
import matplotlib.pyplot as plt
from tqdm import trange
import copy
from .imputation import imputer
from .kernel_class import kernel as ker
from .kernel_class import combine
from sklearn.decomposition import KernelPCA

class dgp:
    """
    Class that contains the deep GP hierarchy for stochastic imputation inference.

    Args:
        X (ndarray): a numpy 2d-array where each row is an input data point and 
            each column is an input dimension. 
        Y (ndarray): a numpy 2d-arrays containing observed output data. 
            The 2d-array has it rows being output data points and columns being output dimensions 
            (with the number of columns equals to the number of GP nodes in the final layer). 
        all_layer (list, optional): a list contains *L* (the number of layers) sub-lists, each of which contains 
            the GPs defined by the kernel class in that layer. The sub-lists are placed in the list 
            in the same order of the specified DGP model. Defaults to `None`. If a DGP structure is not provided, 
            an input-connected two-layered DGP structure (for deterministic model emulation) with the number 
            of GP nodes in the first layer equal to the dimension of **X** is automatically constructed.
        check_rep (bool, optional): whether to check the repetitions in the dataset, i.e., if one input
            position has multiple outputs. Defaults to `True`.
        rff (bool, optional): whether to use random Fourier features to approximate the correlation matrices 
            during the imputation in training. Defaults to `False`.
        M (int, optional): the number of features to be used by random Fourier approximation. It is only used
            when **rff** is set to `True`. Defaults to `None`. If it is not specified, **M** is set to 
            ``max(100, ceil(sqrt(Data Size)*log(Data Size))))``.
        
    Remark:
        This class is used for DGP structures, in which internal I/O are unobservable. When some internal layers
        are fully observable, the DGP model reduces to linked (D)GP model. In such a case, use :class:`.lgp` class for 
        inference where one can have separate input/output training data for each (D)GP. See :class:`.lgp` class for 
        implementation details. 

    Examples:
        To build a list that represents a three-layer DGP with three GPs in the first two layers and
        one GP (i.e., only one dimensional output) in the final layer, do::

            from kernel_class import kernel, combine
            layer1, layer2, layer3=[],[],[]
            for _ in range(3):
                layer1.append(kernel(length=np.array([1])))
            for _ in range(3):
                layer2.append(kernel(length=np.array([1])))
            layer3.append(kernel(length=np.array([1])))
            all_layer=combine(layer1,layer2,layer3) 

    """

    def __init__(self, X, Y, all_layer=None, check_rep=True, rff=False, M=None):
        self.Y=Y
        if isinstance(self.Y, list):
            if len(self.Y)==1:
                self.Y=self.Y[0]
            else:
                raise Exception('Y has to be a numpy 2d-array rather than a list. The list version of Y (for linked emulation) has been reduced. Please use the dedicated lgp class for linked emulation.')
        if (self.Y).ndim==1 or X.ndim==1:
            raise Exception('The input and output data have to be numpy 2d-arrays.')
        #if len(X)>=500 and rff==False:
        #    print('Your training data size is greater than %i, you might want to consider random Fourier features for a faster training by setting "rff=True".' % (500))
        #elif len(X)<=500 and rff:
        #    print('Your training data size is smaller than %i, you may not gain much on the computation by using random Fourier features at the expense of accuracy.' % (500))
        self.indices=None
        if check_rep:
            X0, indices = np.unique(X, return_inverse=True,axis=0)
            if len(X0) != len(X):
                self.X = X0
                self.indices=indices
            else:  
                self.X=X
        else:
            self.X=X
        self.rff=rff
        if M is None:
            self.M=max(100, int(np.ceil(np.sqrt(len(self.X))*np.log(len(self.X)))))
        else:
            self.M=M
        if all_layer is None:
            D, Y_D=np.shape(self.X)[1], np.shape(self.Y)[1]
            layer1 = [ker(length=np.array([1.])) for _ in range(D)]
            layer2 = [ker(length=np.array([1.]),scale_est=True,connect=np.arange(D)) for _ in range(Y_D)]
            all_layer=combine(layer1,layer2)
        self.all_layer=all_layer
        self.n_layer=len(self.all_layer)
        self.initialize()
        self.imp=imputer(self.all_layer)
        (self.imp).sample(burnin=10)
        self.N=0

    def initialize(self):
        """Initialise all_layer attribute for training.
        """
        global_in=self.X
        In=self.X
        for l in range(self.n_layer):
            layer=self.all_layer[l]
            num_kernel=len(layer)
            if l!=self.n_layer-1:
                if np.shape(In)[1]==num_kernel:
                    Out=In
                elif np.shape(In)[1]>num_kernel:
                    pca=KernelPCA(n_components=num_kernel, kernel='sigmoid')
                    Out=pca.fit_transform(In)
                else:
                    Out=np.concatenate((In, In[:,np.random.choice(np.shape(In)[1],num_kernel-np.shape(In)[1])]),1)
                Out=copy.deepcopy(Out)
            for k in range(num_kernel):
                kernel=layer[k]
                if l==self.n_layer-1 and self.indices is not None:
                    kernel.rep=self.indices
                if kernel.input_dim is not None:
                    if l==self.n_layer-1:
                        if kernel.type=='likelihood':
                            if kernel.name=='Poisson' and len(kernel.input_dim)!=1:
                                raise Exception('You need one and only one GP node to feed the ' + kernel.name + ' likelihood node.')
                            elif (kernel.name=='Hetero' or kernel.name=='NegBin') and len(kernel.input_dim)!=2:
                                raise Exception('You need two and only two GP nodes to feed the ' + kernel.name + ' likelihood node.')
                        if kernel.rep is None:
                            kernel.input=copy.deepcopy(In[:,kernel.input_dim])
                        else:
                            kernel.input=copy.deepcopy(In[kernel.rep,:][:,kernel.input_dim])
                    else:
                        kernel.input=copy.deepcopy(In[:,kernel.input_dim])
                else:
                    if l==self.n_layer-1:
                        kernel.input_dim=copy.deepcopy(np.arange(np.shape(In)[1]))
                        if kernel.type=='likelihood':
                            if kernel.name=='Poisson' and len(kernel.input_dim)!=1:
                                raise Exception('You need one and only one GP node to feed the ' + kernel.name + ' likelihood node.')
                            elif (kernel.name=='Hetero' or kernel.name=='NegBin') and len(kernel.input_dim)!=2:
                                raise Exception('You need two and only two GP nodes to feed the ' + kernel.name + ' likelihood node.')
                        if kernel.rep is None:
                            kernel.input=copy.deepcopy(In)
                        else:
                            kernel.input=copy.deepcopy(In[kernel.rep,:])
                    else:
                        kernel.input=copy.deepcopy(In)
                        kernel.input_dim=copy.deepcopy(np.arange(np.shape(In)[1]))
                if kernel.type=='gp':
                    if kernel.connect is not None:
                        if l==self.n_layer-1:
                            if kernel.rep is None:
                                kernel.global_input=copy.deepcopy(global_in[:,kernel.connect])
                            else:
                                kernel.global_input=copy.deepcopy(global_in[kernel.rep,:][:,kernel.connect])
                        else:
                            if l==0 and len(np.intersect1d(kernel.connect,kernel.input_dim))!=0:
                                raise Exception('The local input and global input should not have any overlap. Change input_dim or connect so they do not have any common indices.')
                            kernel.global_input=copy.deepcopy(global_in[:,kernel.connect])
                    kernel.rff, kernel.M = self.rff, self.M
                    kernel.D=np.shape(kernel.input)[1]
                    if kernel.connect is not None:
                        kernel.D+=len(kernel.connect)
                    if kernel.rff:
                        kernel.sample_basis()
                if l==self.n_layer-1:
                    kernel.output=copy.deepcopy(self.Y[:,[k]])
                else:
                    kernel.output=copy.deepcopy(Out[:,k].reshape((-1,1)))
            if l!=self.n_layer-1:
                In=copy.deepcopy(Out)
       
    def train(self, N=500, ess_burn=10, disable=False):
        """Train the DGP model.

        Args:
            N (int): number of iterations for stochastic EM. Defaults to `500`.
            ess_burn (int, optional): number of burnin steps for the ESS-within-Gibbs
                at each I-step of the SEM. Defaults to `10`.
            disable (bool, optional): whether to disable the training progress bar. 
                Defaults to `False`.
        """
        pgb=trange(1,N+1,disable=disable)
        for i in pgb:
            #I-step           
            (self.imp).sample(burnin=ess_burn)
            #M-step
            for l in range(self.n_layer):
                for kernel in self.all_layer[l]:
                    if kernel.type=='gp':
                        kernel.maximise()
                pgb.set_description('Iteration %i: Layer %i' % (i,l+1))
        self.N += N

    def estimate(self,burnin=None):
        """Compute the point estimates of model parameters and output the trained DGP.

        Args:
            burnin (int, optional): the number of SEM iterations to be discarded for
                point estimate calculation. Must be smaller than the SEM iterations 
                implemented. If this is not specified, only the last 25% of iterations
                are used. Defaults to `None`.

        Returns:
            list: an updated list that represents the trained DGP hierarchy.
        """
        if burnin==None:
            burnin=int(self.N*(3/4))
        final_struct=copy.deepcopy(self.all_layer)
        for l in range(len(final_struct)):
            for kernel in final_struct[l]:
                if kernel.type=='gp':
                    point_est=np.mean(kernel.para_path[burnin:,:],axis=0)
                    kernel.scale=point_est[0]
                    kernel.length=point_est[1:-1]
                    kernel.nugget=point_est[-1]
        return final_struct

    def plot(self,layer_no,ker_no,width=4.,height=1.,ticksize=5.,labelsize=8.,hspace=0.1):
        """Plot the traces of model parameters of a particular GP node in the DGP hierarchy.

        Args:
            layer_no (int): the index of the interested layer.
            ker_no (int): the index of the interested GP in the layer specified by **layer_no**.
            width (float, optional): the overall plot width. Defaults to `4`.
            height (float, optional): the overall plot height. Defaults to `1`.
            ticksize (float, optional): the size of sub-plot ticks. Defaults to `5`.
            labelsize (float, optional): the font size of y labels. Defaults to `8`.
            hspace (float, optional): the space between sub-plots. Defaults to `0.1`.
        """
        kernel=self.all_layer[layer_no-1][ker_no-1]
        n_para=np.shape(kernel.para_path)[1]
        fig, axes = plt.subplots(n_para,figsize=(width,n_para*height), dpi= 100,sharex=True)
        fig.tight_layout()
        fig.subplots_adjust(hspace = hspace)
        for p in range(n_para):
            axes[p].plot(kernel.para_path[:,p])
            axes[p].tick_params(axis = 'both', which = 'major', labelsize = ticksize)
            if p==0:
                axes[p].set_ylabel(r'$\sigma^2$',fontsize = labelsize)
            elif p==n_para-1:
                axes[p].set_ylabel(r'$\eta$',fontsize = labelsize)
            else:
                axes[p].set_ylabel(r'$\gamma_{%i}$' %p, fontsize = labelsize)
        plt.show()

    


from numpy.random import uniform
import numpy as np
from .functions import update_f, fmvn

class imputer:
    """Class to implement imputation of latent variables.

    Args:
        all_layer (list): a list that contains the DGP model
    """
    def __init__(self, all_layer):
        self.all_layer=all_layer

    def sample(self,burnin=0):
        """Implement the imputation via the ESS-within-Gibbs.

        Args:
            burnin (int, optional): the number of burnin iterations for the ESS-within-Gibbs sampler
                to generate one realisation of latent variables. Defaults to `0`.
        """
        n_layer=len(self.all_layer)
        for _ in range(burnin+1):
            for l in range(n_layer-1):
                layer=self.all_layer[l]
                n_kernel=len(layer)
                for k in range(n_kernel):
                    target_kernel=layer[k]
                    linked_upper_kernels=[kernel for kernel in self.all_layer[l+1] if k in kernel.input_dim]
                    self.one_sample(target_kernel,linked_upper_kernels,k)
    
    @staticmethod
    def one_sample(target_kernel,linked_upper_kernels,k):
        """Impute one latent variable produced by a particular GP.

        Args:
            target_kernel (class): the GP whose output is a latent variable that needs to be imputed.
            linked_upper_kernels (list): a list of GPs (in the next layer) that link the output produced
                by the GP defined by the argument **target_kernel**.
            k (int): the index indicating the position of the GP defined by the argument **target_kernel** in
                its layer.
        """
        covariance=target_kernel.k_matrix()
        covariance=target_kernel.scale*covariance
                  
        if len(linked_upper_kernels)==1 and linked_upper_kernels[0].type=='likelihood' and linked_upper_kernels[0].exact_post_idx!=None:
            idx=np.where(linked_upper_kernels[0].input_dim == k)[0]
            if idx in linked_upper_kernels[0].exact_post_idx:
                f=linked_upper_kernels[0].posterior(idx=idx,v=covariance)
                if linked_upper_kernels[0].rep is None:
                    linked_upper_kernels[0].input[:,idx]=f.reshape(-1,1)
                else:
                    linked_upper_kernels[0].input[:,idx]=f[linked_upper_kernels[0].rep].reshape(-1,1)
                target_kernel.output[:,0]=f
                return
        
        f=(target_kernel.output).flatten()
        # Choose the ellipse for this sampling iteration.
        #nu = np.random.default_rng().multivariate_normal(mean=np.zeros(len(f)),cov=covariance,check_valid='ignore')        
        nu = fmvn(covariance)               
        # Set the candidate acceptance threshold.
        log_y=0
        for linked_kernel in linked_upper_kernels:
            if linked_kernel.type=='gp':
                if linked_kernel.rff:
                    log_y += linked_kernel.log_likelihood_func_rff()
                else:      
                    log_y += linked_kernel.log_likelihood_func()
            elif linked_kernel.type=='likelihood': 
                log_y += linked_kernel.llik()
        log_y += np.log(uniform())
        # Set the bracket for selecting candidates on the ellipse.
        theta = uniform(0., 2.*np.pi)
        theta_min, theta_max = theta - 2.*np.pi, theta

        # Iterates until an candidate is selected.
        while True:
            # Generates a point on the ellipse defines by `nu` and the input. We
            # also compute the log-likelihood of the candidate and compare to
            # our threshold.
            fp = update_f(f,nu,theta)
            log_yp=0
            for linked_kernel in linked_upper_kernels:
                if linked_kernel.rep is None:
                    linked_kernel.input[:,linked_kernel.input_dim==k]=fp.reshape(-1,1)
                else:
                    linked_kernel.input[:,linked_kernel.input_dim==k]=fp[linked_kernel.rep].reshape(-1,1)
                if linked_kernel.type=='gp':
                    if linked_kernel.rff:
                        log_yp += linked_kernel.log_likelihood_func_rff()
                    else:
                        log_yp += linked_kernel.log_likelihood_func()
                elif linked_kernel.type=='likelihood': 
                    log_yp += linked_kernel.llik()
            if log_yp > log_y:
                target_kernel.output[:,0]=fp
                return
            else:
                # If the candidate is not selected, shrink the bracket and
                # generate a new `theta`, which will yield a new candidate
                # point on the ellipse.
                if theta < 0.:
                    theta_min = theta
                else:
                    theta_max = theta
                theta = uniform(theta_min, theta_max)
    
    def key_stats(self):
        """Compute and store key statistics used in predictions
        """
        n_layer=len(self.all_layer)
        for l in range(n_layer):
            layer=self.all_layer[l]
            for kernel in layer:
                if kernel.type == 'gp':
                    kernel.compute_stats()



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

VERSION = '2.1.1'
PACKAGE_NAME = 'dgpsi'
AUTHOR = 'Deyu Ming'
AUTHOR_EMAIL = 'deyu.ming.16@ucl.ac.uk'
URL = 'https://github.com/mingdeyu/DGP'

LICENSE = 'MIT'
DESCRIPTION = 'Deep and Linked Gaussian Process Emulations using Stochastic Imputation'
LONG_DESCRIPTION = (here / 'README.md').read_text(encoding='utf-8')

INSTALL_REQUIRES = [
      'numpy>=1.18.2',
      'numba>=0.51.2',
      'matplotlib>=3.2.1',
      'tqdm>=4.50.2',
      'scipy>=1.4.1',
      'scikit-learn>=0.22.0',
      'jupyter>=1.0.0',
      'dill>=0.3.2',
      'pathos>=0.2.9',
      'psutil>=5.8.0',
      'cython>=0.29.30',
      'pybind11>=2.10.0',
      'pythran>=0.11.0',
      'scikit-build>=0.15.0',
      'tabulate>=0.8.7'
]

setup(name=PACKAGE_NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type='text/markdown',
      author=AUTHOR,
      license=LICENSE,
      keywords = ["surrogate modelling","deep learning","stochastic EM","elliptical slice sampling"],
      author_email=AUTHOR_EMAIL,
      url=URL,
      install_requires=INSTALL_REQUIRES,
      python_requires='>=3.7, <4',
      extras_require={'interactive': ['jupyter']},
      packages=find_packages(include=['dgpsi', 'dgpsi.*'])
      )


from dill import dump, load
from tabulate import tabulate
import numpy as np

######Save and Load Emulators#######
def write(emu, pkl_file):
    """Save the constructed emulator to a pkl file.
    
    Args:
        emu (class): an emulator class. For GP, it is the :class:`.gp` class after training. 
            For DGP, it is the :class:`.emulator` class. For linked GP/DGP, it is the :class:`.lgp` class.
        pkl_file (strings): specifies the path to and the name of the `.pkl` file to which
            the emulator is saved.
    """
    dump(emu, open(pkl_file+".pkl","wb"))


def read(pkl_file):
    """Load the `.pkl` file that stores the emulator.
    
    Args:
        pkl_file (strings): specifies the path to and the name of the `.pkl` file where
            the emulator is stored.
    
    Returns:
        class: an emulator class. For GP, it is the :class:`.gp` class. For DGP, it is the :class:`.emulator` class. 
        For linked GP/DGP, it is the :class:`.lgp` class.
    """
    emu = load(open(pkl_file+".pkl", "rb"))
    return emu

######summary function#######
def summary(obj):
    """Summarize key information of GP, DGP, and Linked (D)GP structures.

    Args:
        obj (class): **obj** can be one of the following:

            1. an instance of :class:`.kernel` class;
            2. an instance of :class:`.gp` class;
            3. an instance of :class:`.dgp` class;
            4. an instance of :class:`.emulator` class;
            5. an instance of :class:`.lgp` class
    
    Returns:
        string: a table summarizing key information contained in **obj**.
    """
    info=[]
    if type(obj).__name__=='kernel':
        info.append(['Kernel Fun', 'Length-scale(s)', 'Variance', 'Nugget'])
        info.append(['Squared-Exp' if obj.name=='sexp' else 'Matern-2.5',
        f"{np.array2string(obj.length, precision=3, floatmode='fixed', separator=', ')}", 
        f"{np.array2string(np.atleast_1d(obj.scale)[0], precision=3, floatmode='fixed')}" if obj.scale_est else f"{np.array2string(np.atleast_1d(obj.scale)[0], precision=3, floatmode='fixed')} (fixed)", 
        f"{np.array2string(np.atleast_1d(obj.nugget)[0], precision=3, floatmode='fixed')}" if obj.nugget_est else f"{np.array2string(np.atleast_1d(obj.nugget)[0], precision=3, floatmode='fixed')} (fixed)",
        ])
        table = tabulate(info, headers='firstrow', tablefmt='fancy_grid')
        print(table)
    elif type(obj).__name__=='gp':
        ker=obj.kernel
        info.append(['Kernel Fun', 'Length-scale(s)', 'Variance', 'Nugget', 'Input Dims'])
        info.append(['Squared-Exp' if ker.name=='sexp' else 'Matern-2.5',
        f"{np.array2string(ker.length, precision=3, floatmode='fixed', separator=', ')}", 
        f"{np.array2string(np.atleast_1d(ker.scale)[0], precision=3, floatmode='fixed')}" if ker.scale_est else f"{np.array2string(np.atleast_1d(ker.scale)[0], precision=3, floatmode='fixed')} (fixed)", 
        f"{np.array2string(np.atleast_1d(ker.nugget)[0], precision=3, floatmode='fixed')}" if ker.nugget_est else f"{np.array2string(np.atleast_1d(ker.nugget)[0], precision=3, floatmode='fixed')} (fixed)",
        f"{np.array2string(ker.input_dim+1, separator=', ')}"])
        table = tabulate(info, headers='firstrow', tablefmt='fancy_grid')
        print(table)
        print("'Input Dims' indicates the dimensions (i.e., columns) of your input data that are actually used for GP training.")
    elif type(obj).__name__=='dgp':
        if obj.N!=0:
            print('To get the summary of the trained DGP model, construct an emulator instance using the emulator() class and then apply summary() to it.')
            return
        all_layer = obj.all_layer
        info.append(['Layer No.', 'Node No.', 'Type', 'Length-scale(s)', 'Variance', 'Nugget', 'Input Dims', 'Global Connection'])
        for l in range(obj.n_layer):
            layer=all_layer[l]
            for k in range(len(layer)):
                ker=layer[k]
                info.append([f'Layer {l+1:d}', f'Node {k+1:d}',
                    'GP (Squared-Exp)' if ker.name=='sexp' else 'GP (Matern-2.5)' if ker.name=='matern2.5' else f'Likelihood ({ker.name})',
                    'NA' if ker.type=='likelihood' else f"{np.array2string(ker.length, precision=3, floatmode='fixed', separator=', ')}", 
                    'NA' if ker.type=='likelihood' else f"{np.array2string(np.atleast_1d(ker.scale)[0], precision=3, floatmode='fixed')}" if ker.scale_est else f"{np.array2string(np.atleast_1d(ker.scale)[0], precision=3, floatmode='fixed')} (fixed)", 
                    'NA' if ker.type=='likelihood' else f"{np.array2string(np.atleast_1d(ker.nugget)[0], precision=3, floatmode='fixed')}" if ker.nugget_est else f"{np.array2string(np.atleast_1d(ker.nugget)[0], precision=3, floatmode='fixed')} (fixed)",
                    f"{np.array2string(ker.input_dim+1, separator=', ')}",
                    'NA' if ker.type=='likelihood' else f"{np.array2string(ker.connect+1, separator=', ')}" if ker.connect is not None else 'No'])
        table = tabulate(info, headers='firstrow', tablefmt='fancy_grid')
        print(table)
        print("1. 'Input Dims' presents the indices of GP nodes in the feeding layer whose outputs are used as the input to the current GP.")
        print("2. 'Global Connection' indicates the dimensions (i.e., column numbers) of the global input data that are used as additional input dimensions to the current GP.")
    elif type(obj).__name__=='emulator':
        all_layer = obj.all_layer
        info.append(['Layer No.', 'Node No.', 'Type', 'Length-scale(s)', 'Variance', 'Nugget', 'Input Dims', 'Global Connection'])
        for l in range(obj.n_layer):
            layer=all_layer[l]
            for k in range(len(layer)):
                ker=layer[k]
                info.append([f'Layer {l+1:d}', f'Node {k+1:d}',
                    'GP (Squared-Exp)' if ker.name=='sexp' else 'GP (Matern-2.5)' if ker.name=='matern2.5' else f'Likelihood ({ker.name})',
                    'NA' if ker.type=='likelihood' else f"{np.array2string(ker.length, precision=3, floatmode='fixed', separator=', ')}", 
                    'NA' if ker.type=='likelihood' else f"{np.array2string(np.atleast_1d(ker.scale)[0], precision=3, floatmode='fixed')}" if ker.scale_est else f"{np.array2string(np.atleast_1d(ker.scale)[0], precision=3, floatmode='fixed')} (fixed)", 
                    'NA' if ker.type=='likelihood' else f"{np.array2string(np.atleast_1d(ker.nugget)[0], precision=3, floatmode='fixed')}" if ker.nugget_est else f"{np.array2string(np.atleast_1d(ker.nugget)[0], precision=3, floatmode='fixed')} (fixed)",
                    f"{np.array2string(ker.input_dim+1, separator=', ')}",
                    'NA' if ker.type=='likelihood' else f"{np.array2string(ker.connect+1, separator=', ')}" if ker.connect is not None else 'No'])
        table = tabulate(info, headers='firstrow', tablefmt='fancy_grid')
        print(table)
        print("1. 'Input Dims' presents the indices of GP nodes in the feeding layer whose outputs are used as the input to the current GP.")
        print("2. 'Global Connection' indicates the dimensions (i.e., column numbers) of the global input data that are used as additional input dimensions to the current GP.")
    elif type(obj).__name__=='lgp':
        all_layer = obj.all_layer
        info.append(['Layer No.', 'Emulator No.', 'Type', 'Connection', 'External Inputs'])
        for l in range(obj.L):
            layer=all_layer[l]
            for k in range(len(layer)):
                cont=layer[k]
                if l==0:
                    links = f"Global input: {np.array2string(cont.local_input_idx+1, separator=', ')}"
                    external = 'No'
                else:
                    emu_idx, output_idx, emu_count = [], [], 0
                    for feeding_cont in all_layer[l-1]:
                        n = 1 if feeding_cont.type=='gp' else len(feeding_cont.structure[-1])
                        emu_idx, output_idx = np.concatenate((emu_idx, np.array([emu_count]*n))), np.concatenate((output_idx, np.arange(n)))
                        emu_count += 1
                    connected_emu, connected_output = emu_idx[cont.local_input_idx], output_idx[cont.local_input_idx]
                    links = ''
                    for i in range(len(cont.local_input_idx)):
                        links += f"Emu {np.int64(connected_emu[i]+1)} in Layer {l}: output {np.int64(connected_output[i]+1)}\n"
                    if cont.type == 'gp':
                        external = 'No' if cont.structure.connect is None else 'Yes'
                    else:
                        external = 'No' if cont.structure[0][0].connect is None else 'Yes'
                info.append([f'Layer {l+1:d}', f'Emu {k+1:d}', 'DGP' if cont.type=='dgp' else 'GP', links, external
                    ])
        table = tabulate(info, headers='firstrow', tablefmt='fancy_grid')
        print(table)
        print("1. 'Connection' gives the indices of emulators and the associated output dimensions that are linked to the current emulator.")
        print("2. 'External Inputs' indicates if the current emulator has external inputs that are not provided by the feeding emulators.")